"""
.. currentmodule:: dgl

Graph Classification Tutorial
=============================

**Author**: `Mufei Li <https://github.com/mufeili>`_,
`Minjie Wang <https://jermainewang.github.io/>`_,
`Zheng Zhang <https://shanghai.nyu.edu/academics/faculty/directory/zheng-zhang>`_.

In this tutorial, you learn how to use DGL to batch multiple graphs of variable size and shape. The
tutorial also demonstrates training a graph neural network for a simple graph classification task.

Graph classification is an important problem
with applications across many fields, such as bioinformatics, chemoinformatics, social
network analysis, urban computing, and cybersecurity. Applying graph neural
networks to this problem has been a popular approach recently. This can be seen in the following reserach references:
`Ying et al., 2018 <https://arxiv.org/abs/1806.08804>`_,
`Cangea et al., 2018 <https://arxiv.org/abs/1811.01287>`_,
`Knyazev et al., 2018 <https://arxiv.org/abs/1811.09595>`_,
`Bianchi et al., 2019 <https://arxiv.org/abs/1901.01343>`_,
`Liao et al., 2019 <https://arxiv.org/abs/1901.01484>`_,
`Gao et al., 2019 <https://openreview.net/forum?id=HJePRoAct7>`_).

"""

###############################################################################
# Simple graph classification task
# --------------------------------
# In this tutorial, you learn how to perform batched graph classification
# with DGL. The example task objective is to classify eight types of topologies shown here.
#
# .. image:: https://data.dgl.ai/tutorial/batch/dataset_overview.png
#     :align: center
#
# Implement a synthetic dataset :class:`data.MiniGCDataset` in DGL. The dataset has eight
# different types of graphs and each class has the same number of graph samples.
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader
import dgl
from dgl.data import MiniGCDataset
# import matplotlib.pyplot as plt
# import networkx as nx

from ...data.data import GraphData
from ...modules.graph_embedding.gat import GAT
from ...modules.prediction.classification.graph_classification import FeedForwardNN




# fig, ax = plt.subplots()
# nx.draw(graph.to_networkx(), ax=ax)
# ax.set_title('Class: {:d}'.format(label))
# plt.show()

###############################################################################
# Form a graph mini-batch
# -----------------------
# To train neural networks efficiently, a common practice is to batch
# multiple samples together to form a mini-batch. Batching fixed-shaped tensor
# inputs is common. For example, batching two images of size 28 x 28
# gives a tensor of shape 2 x 28 x 28. By contrast, batching graph inputs
# has two challenges:
#
# * Graphs are sparse.
# * Graphs can have various length. For example, number of nodes and edges.
#
# To address this, DGL provides a :func:`dgl.batch` API. It leverages the idea that
# a batch of graphs can be viewed as a large graph that has many disjointed
# connected components. Below is a visualization that gives the general idea.
#
# .. image:: https://data.dgl.ai/tutorial/batch/batch.png
#     :width: 400pt
#     :align: center
#
# Define the following ``collate`` function to form a mini-batch from a given
# list of graph and label pairs.



def collate(samples):
    # The input `samples` is a list of pairs
    #  (graph, label).
    graphs, labels = map(list, zip(*samples))
    batched_graph = dgl.batch(graphs)
    return batched_graph, torch.tensor(labels)

###############################################################################
# The return type of :func:`dgl.batch` is still a graph. In the same way,
# a batch of tensors is still a tensor. This means that any code that works
# for one graph immediately works for a batch of graphs. More importantly,
# because DGL processes messages on all nodes and edges in parallel, this greatly
# improves efficiency.
#
# Graph classifier
# ----------------
# Graph classification proceeds as follows.
#
# .. image:: https://data.dgl.ai/tutorial/batch/graph_classifier.png
#
# From a batch of graphs, perform message passing and graph convolution
# for nodes to communicate with others. After message passing, compute a
# tensor for graph representation from node (and edge) attributes. This step might
# be called readout or aggregation. Finally, the graph
# representations are fed into a classifier :math:`g` to predict the graph labels.
#
# Graph convolution layer can be found in the ``dgl.nn.<backend>`` submodule.


###############################################################################
# Readout and classification
# --------------------------
# For this demonstration, consider initial node features to be their degrees.
# After two rounds of graph convolution, perform a graph readout by averaging
# over all node features for each graph in the batch.
#
# .. math::
#
#    h_g=\frac{1}{|\mathcal{V}|}\sum_{v\in\mathcal{V}}h_{v}
#
# In DGL, :func:`dgl.mean_nodes` handles this task for a batch of
# graphs with variable size. You then feed the graph representations into a
# classifier with one linear layer to obtain pre-softmax logits.




class GNNClassifier(nn.Module):
    def __init__(self,
                num_layers,
                input_size,
                hidden_size,
                output_size,
                num_heads,
                num_out_heads,
                direction_option,
                graph_pooling,
                feat_drop=0.6,
                attn_drop=0.6,
                negative_slope=0.2,
                residual=False,
                activation=F.elu):
        super(GNNClassifier, self).__init__()
        self.direction_option = direction_option
        heads = [num_heads] * (num_layers - 1) + [num_out_heads]
        self.gnn = GAT(num_layers,
                    input_size,
                    hidden_size,
                    hidden_size,
                    heads,
                    direction_option=direction_option,
                    feat_drop=feat_drop,
                    attn_drop=attn_drop,
                    negative_slope=negative_slope,
                    residual=residual,
                    activation=activation)

        self.clf = FeedForwardNN(2 * hidden_size if self.direction_option == 'bi_sep' else hidden_size,
                        output_size,
                        [hidden_size],
                        graph_pool_type=graph_pooling)


    def forward(self, batched_graph):
        # Use node degree as the initial node feature. For undirected graphs, the in-degree
        # is the same as the out_degree.
        node_feat = batched_graph.in_degrees().view(-1, 1).float()
        batched_graph.ndata['node_feat'] = node_feat

        batched_graph_data = GraphData()
        batched_graph_data.from_dgl(batched_graph)
        batched_graph_data = self.gnn(batched_graph_data)

        batched_graph.ndata['node_emb'] = batched_graph_data.node_features['node_emb']

        batched_graph = self.clf(batched_graph)
        logits = batched_graph.graph_attributes['logits']

        return logits


###############################################################################
# Setup and training
# ------------------
# Create a synthetic dataset of :math:`400` graphs with :math:`10` ~
# :math:`20` nodes. :math:`320` graphs constitute a training set and
# :math:`80` graphs constitute a test set.

def main(args):
    # A dataset with 80 samples, each graph is
    # of size [10, 20]
    dataset = MiniGCDataset(80, 10, 20)
    graph, label = dataset[0]

    # Create training and test sets.
    trainset = MiniGCDataset(320, 10, 20)
    testset = MiniGCDataset(80, 10, 20)
    # Use PyTorch's DataLoader and the collate function
    # defined before.
    data_loader = DataLoader(trainset, batch_size=args.batch_size, shuffle=True,
                             collate_fn=collate)

    num_feats = 1
    num_classes = 8
    # Create model
    model = GNNClassifier(args.num_layers,
                    num_feats,
                    args.num_hidden,
                    num_classes,
                    args.num_heads,
                    args.num_out_heads,
                    direction_option=args.direction_option,
                    graph_pooling=args.graph_pooling,
                    feat_drop=args.in_drop,
                    attn_drop=args.attn_drop,
                    negative_slope=args.negative_slope,
                    residual=args.residual,
                    activation=F.elu)


    loss_func = nn.CrossEntropyLoss()
    optimizer = optim.Adam(model.parameters(), lr=args.lr)
    model.train()

    epoch_losses = []
    for epoch in range(args.epochs):
        epoch_loss = 0
        for iter, (bg, label) in enumerate(data_loader):
            prediction = model(bg)
            loss = loss_func(prediction, label)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            epoch_loss += loss.detach().item()
        epoch_loss /= (iter + 1)
        print('Epoch {}, loss {:.4f}'.format(epoch, epoch_loss))
        epoch_losses.append(epoch_loss)

    ###############################################################################
    # The learning curve of a run is presented below.

    # plt.title('cross entropy averaged over minibatches')
    # plt.plot(epoch_losses)
    # plt.show()

    ###############################################################################
    # The trained model is evaluated on the test set created. To deploy
    # the tutorial, restrict the running time to get a higher
    # accuracy (:math:`80` % ~ :math:`90` %) than the ones printed below.

    model.eval()
    # Convert a list of tuples to two lists
    test_X, test_Y = map(list, zip(*testset))
    test_bg = dgl.batch(test_X)
    test_Y = torch.tensor(test_Y).float().view(-1, 1)
    probs_Y = torch.softmax(model(test_bg), 1)
    sampled_Y = torch.multinomial(probs_Y, 1)
    argmax_Y = torch.max(probs_Y, 1)[1].view(-1, 1)
    print('Accuracy of sampled predictions on the test set: {:.4f}%'.format(
        (test_Y == sampled_Y.float()).sum().item() / len(test_Y) * 100))
    print('Accuracy of argmax predictions on the test set: {:4f}%'.format(
        (test_Y == argmax_Y.float()).sum().item() / len(test_Y) * 100))


if __name__ == '__main__':
    # Training settings
    parser = argparse.ArgumentParser(description='GAT')
    # register_data_args(parser)
    parser.add_argument("--graph-pooling", type=str, default='max_pool',
                        help="graph pooling (`avg_pool`, `max_pool`)")
    parser.add_argument("--direction-option", type=str, default='uni',
                        help="direction type (`uni`, `bi_fuse`, `bi_sep`)")
    parser.add_argument("--num-heads", type=int, default=8,
                        help="number of hidden attention heads")
    parser.add_argument("--num-out-heads", type=int, default=1,
                        help="number of output attention heads")
    parser.add_argument("--num-layers", type=int, default=2,
                        help="number of hidden layers")
    parser.add_argument("--num-hidden", type=int, default=8,
                        help="number of hidden units")
    parser.add_argument("--residual", action="store_true", default=False,
                        help="use residual connection")
    parser.add_argument("--in-drop", type=float, default=.6,
                        help="input feature dropout")
    parser.add_argument("--attn-drop", type=float, default=.6,
                        help="attention dropout")
    parser.add_argument("--lr", type=float, default=0.001,
                        help="learning rate")
    parser.add_argument('--weight-decay', type=float, default=5e-4,
                        help="weight decay")
    parser.add_argument('--negative-slope', type=float, default=0.2,
                        help="the negative slope of leaky relu")
    parser.add_argument('--early-stop', action='store_true', default=False,
                        help="indicates whether to use early stop or not")
    parser.add_argument("--patience", type=int, default=100,
                        help="early stopping patience")
    parser.add_argument('--device', type=int, default=0,
                        help='which gpu to use if any (default: 0)')
    parser.add_argument('--drop_ratio', type=float, default=0.5,
                        help='dropout ratio (default: 0.5)')
    parser.add_argument('--batch-size', type=int, default=32,
                        help='input batch size for training (default: 32)')
    parser.add_argument('--epochs', type=int, default=80,
                        help='number of epochs to train (default: 100)')
    parser.add_argument('--num_workers', type=int, default=0,
                        help='number of workers (default: 0)')
    parser.add_argument('--dataset', type=str, default="ogbg-molhiv",
                        help='dataset name (default: ogbg-molhiv)')
    parser.add_argument('--feature', type=str, default="full",
                        help='full feature or simple feature')
    parser.add_argument('--filename', type=str, default="",
                        help='filename to output result (default: )')
    args = parser.parse_args()

    main(args)



import abc
import torch.nn as nn


class GNNLayerBase(nn.Module):
    def __init__(self):
        super(GNNLayerBase, self).__init__()

    @abc.abstractmethod
    def forward(self, graph, node_feat):
        raise NotImplementedError('GNNLayerBase: Not Implemented.')

class GNNBase(nn.Module):
    def __init__(self):
        super(GNNBase, self).__init__()

    @abc.abstractmethod
    def forward(self, graph, node_feat):
        raise NotImplementedError('GNNBase: Not Implemented.')



import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.optim as optim
import dgl

from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from graph4nlp.pytorch.test.example.syntax_nmt.dataset import EuroparlNMTDataset
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction

class Graph2seq(nn.Module):
    def __init__(self, vocab, hidden_size=300, direction_option='bi_sep'):
        super(Graph2seq, self).__init__()

        self.vocab = vocab
        embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
                           'seq_info_encode_strategy': "bilstm"}
        self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
                                                               vocab=vocab.in_word_vocab,
                                                               hidden_size=hidden_size, dropout=0.2, use_cuda=True,
                                                               fix_word_emb=False)
        self.gnn = None
        self.word_emb = self.graph_topology.embedding_layer.word_emb_layers[0].word_emb_layer
        self.gnn_encoder = GAT(2, hidden_size, hidden_size, hidden_size, [2, 1], direction_option=direction_option)
        self.seq_decoder = StdRNNDecoder(max_decoder_step=50,
                                         decoder_input_size=2 * hidden_size if direction_option == 'bi_sep' else hidden_size,
                                         decoder_hidden_size=hidden_size,
                                         word_emb=self.word_emb, vocab=self.vocab.in_word_vocab,
                                         attention_type="sep_diff_encoder_type", fuse_strategy="concatenate",
                                         rnn_emb_input_size=hidden_size,
                                         tgt_emb_as_output_layer=True, device=self.graph_topology.device)
        self.loss_calc = Graph2seqLoss(self.vocab.in_word_vocab)

    def forward(self, graph_list, tgt=None, require_loss=True):
        batch_dgl_graph = self.graph_topology(graph_list)
        # do graph nn here
        # convert DGLGraph to GraphData
        batch_graph = GraphData()
        batch_graph.from_dgl(batch_dgl_graph)

        # run GNN
        batch_graph = self.gnn_encoder(batch_graph)
        batch_dgl_graph.ndata['node_emb'] = batch_graph.node_features['node_emb']
        batch_dgl_graph.ndata['rnn_emb'] = batch_graph.node_features['node_feat']

        dgl_graph_list = dgl.unbatch(batch_dgl_graph)
        for g, dg in zip(graph_list, dgl_graph_list):
            g.node_features["node_emb"] = dg.ndata["node_emb"]
            g.node_features["rnn_emb"] = dg.ndata["rnn_emb"]

        # down-task
        prob, _, _ = self.seq_decoder(graph_list, tgt_seq=tgt)
        if require_loss:
            loss = self.loss_calc(prob, tgt)
            return prob, loss
        else:
            return prob


class NMT:
    def __init__(self):
        super(NMT, self).__init__()
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        dataset = EuroparlNMTDataset(root_dir="/home/shiina/shiina/lib/dataset",
                                 topology_builder=DependencyBasedGraphConstruction,
                                 topology_subdir='DependencyGraph')
        data_size = len(dataset)
        self.train_dataloader = DataLoader(dataset[:int(0.8 * data_size)], batch_size=24, shuffle=True,
                                           num_workers=10,
                                           collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset[int(0.8 * data_size):], batch_size=24, shuffle=False,
                                          num_workers=10,
                                          collate_fn=dataset.collate_fn)
        self.vocab = dataset.vocab_model

    def _build_model(self):
        self.model = Graph2seq(self.vocab).cuda()

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=1e-3)

    def _build_evaluation(self):
        self.metrics = []

    def train(self):
        max_score = -1
        for epoch in range(200):
            self.model.train()
            print("Epoch: {}".format(epoch))
            for data in self.train_dataloader:
                graph_list, tgt = data
                tgt = tgt.cuda()
                _, loss = self.model(graph_list, tgt, require_loss=True)
                print(loss)
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
            if epoch > 4:
                score = self.evaluate()
                max_score = max(max_score, score)
        return max_score

    def evaluate(self):
        self.model.eval()
        pred_collect = []
        gt_collect = []
        for data in self.test_dataloader:
            graph_list, tgt = data
            prob = self.model(graph_list, require_loss=False)
            pred = logits2seq(prob)

            pred_str = wordid2str(pred.detach().cpu(), self.vocab.in_word_vocab)
            tgt_str = wordid2str(tgt, self.vocab.in_word_vocab)
            pred_collect.extend(pred_str)
            gt_collect.extend(tgt_str)

        score = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
        print("accuracy: {:.3f}".format(score))
        return score




import torch


class SizeMismatchException(Exception):
    pass


class NodeNotFoundException(Exception):
    pass


class EdgeNotFoundException(Exception):
    pass


def int_to_list(x: int or list):
    assert isinstance(x, list) or isinstance(x, int)
    return x if isinstance(x, list) else [x]


def check_and_expand(x: list, y: list):
    assert isinstance(x, list) and isinstance(y, list)
    max_len = max(len(x), len(y))
    if len(x) == len(y):
        return x, y
    elif len(x) * len(y) == max_len:  # Which means at least one of the list is of length 1
        if len(x) == 1:
            x = x * max_len
        else:
            y = y * max_len
        return x, y
    else:
        raise ValueError(
            'The two lists {} and {} cannot be automatically broadcasted to the same length.'.format(x, y))


def slice_to_list(sl, max_len):
    """
    Turn a slice object into a list

    Parameters
    ----------
    sl: slice
        The slice object

    max_len: int
        Max length of the iterable

    Returns
    -------
    list
        The converted list
    """

    if sl.start is None:
        start = 0
    elif sl.start < 0:
        start = max_len + sl.start
    else:
        start = sl.start

    if sl.stop is None:
        stop = max_len
    elif sl.stop < 0:
        stop = max_len + sl.stop
    else:
        stop = sl.stop

    if sl.step is None:
        step = 1
    else:
        step = sl.step

    return list(range(start, stop, step))


def entail_zero_padding(old_tensor: torch.Tensor, num_rows: int):
    if old_tensor is None:
        return None

    if len(old_tensor.shape) == 1:
        return torch.cat((old_tensor, torch.zeros(num_rows).to(dtype=old_tensor.dtype, device=old_tensor.device)))
    else:
        return torch.cat((old_tensor, torch.zeros((num_rows, *old_tensor.shape[1:])).to(dtype=old_tensor.dtype,
                                                                                device=old_tensor.device)), dim=0)


def reverse_index(l: list, v):
    """
    Find the index of the last occurrence of an element in a list.

    Parameters
    ----------
    l: list
        The container of all elements
    v: object
        The element to be found

    Returns
    -------
    int:
        The index of the last occurrence of `v` in `l`.

    Raises
    ------
    ValueError
        If the element is not found in the list.
    """
    if v not in l:
        raise ValueError
    return len(l) - l[::-1].index(v) - 1



import pytest
import torch
import torch.nn as nn

from ...data.data import GraphData, from_batch, to_batch, from_dgl
from ...data.utils import EdgeNotFoundException


def fail_here():
    raise Exception('The above line of code shouldn\'t be executed normally')


def test_add_nodes():
    g = GraphData()
    try:
        g.add_nodes(-1)
        fail_here()
    except AssertionError:
        pass

    g.add_nodes(10)
    assert g.get_node_num() == 10
    assert g.node_features['node_feat'] is None
    assert g.node_features['node_emb'] is None
    assert g.node_attributes[0]['node_attr'] is None

    g.node_features['zero'] = torch.zeros(10)
    g.add_nodes(9)
    assert g.get_node_num() == 19
    assert torch.all(torch.eq(g.node_features['zero'][10:], torch.zeros(9)))


def test_set_node_features():
    g = GraphData()
    g.add_nodes(10)
    # Test adding node features at whole graph level.
    try:
        g.node_features['node_feat'] = torch.randn((9, 10))
        fail_here()
    except AssertionError:
        pass

    g.node_features['node_feat'] = torch.randn((10, 10))
    g.node_features['zero'] = torch.zeros(10)
    g.node_features['idx'] = torch.tensor(list(range(10)), dtype=torch.long)

    for i in range(g.get_node_num()):
        assert g.node_features['idx'][i] == i
        assert g.node_features['node_feat'][i].shape == torch.Size([10])

    # Test modifying node features partially
    g.node_features['zero'][:5] = torch.ones(5)
    for i in range(5):
        assert g.node_features['zero'][i] == 1

    # Test computational graph consistency when modifying node features both partially and completely
    embedding_layer = nn.Embedding(num_embeddings=100, embedding_dim=10)
    optimizer = torch.optim.Adam(params=embedding_layer.parameters())
    target = torch.ones(10)
    init_loss = - torch.nn.functional.cosine_similarity(embedding_layer(g.node_features['idx']).sum(dim=-1), target,
                                                        dim=0)
    for i in range(100):
        g.node_features['node_emb'] = embedding_layer(g.node_features['idx'])
        node_emb_sum = g.node_features['node_emb'].sum(dim=-1)
        loss = - torch.nn.functional.cosine_similarity(node_emb_sum, target, dim=0)
        # print('Epoch {}: loss = {}.'.format(i + 1, loss))
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
    print('Initial loss = {}. Final loss = {}'.format(init_loss, loss))
    assert init_loss > loss


def test_add_edges():
    g = GraphData()
    g.add_nodes(10)

    # Test wrong cases where one of the `src` and `tgt` is empty.
    try:
        g.add_edges([], [0, 1, 2])
        fail_here()
    except ValueError:
        pass

    try:
        g.add_edges([0, 1, 2], [])
        fail_here()
    except ValueError:
        pass

    # Test `src` and `tgt` size mismatch situation.
    try:
        g.add_edges([0, 1, 2], [1, 2])
        fail_here()
    except ValueError:
        pass

    # Test list autocomplete
    g.add_edges([0], [1, 2])
    g.add_edges([3, 4], [5])
    assert g.get_all_edges() == [(0, 1), (0, 2), (3, 5), (4, 5)]

    # Test normal add_edges
    g = GraphData()
    g.add_nodes(10)
    g.add_edges([0, 1, 2, 3, 4], [1, 2, 3, 4, 5])
    assert g.get_all_edges() == [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]

    # Test adding duplicate edges
    with pytest.warns(Warning):
        g.add_edges([0, 1], [1, 2])


def test_edge_ids():
    g = GraphData()
    g.add_nodes(10)
    for i in range(10):
        g.add_edge(src=i, tgt=(i + 1) % 10)
    # all edges are in the graph
    assert g.edge_ids([0, 1, 2, 3, 4], [1, 2, 3, 4, 5]) == [0, 1, 2, 3, 4]
    # some edges are missing
    try:
        g.edge_ids([0, 1, 2], [1, 2, 4])
        fail_here()
    except EdgeNotFoundException:
        pass
    # broadcasting
    g.add_edges(0, [2, 3, 4, 5, 6])
    assert g.edge_ids(0, [1, 2, 3, 4, 5, 6]) == [0, 10, 11, 12, 13, 14]


def test_edge_features():
    g = GraphData()
    g.add_nodes(10)
    for i in range(10):
        g.add_edge(src=i, tgt=(i + 1) % 10)
    g.edge_features['idx'] = torch.tensor(list(range(10)), dtype=torch.long)
    # Test computational graph consistency when modifying node features both partially and completely
    embedding_layer = nn.Embedding(num_embeddings=100, embedding_dim=10)
    optimizer = torch.optim.Adam(params=embedding_layer.parameters())
    target = torch.ones(10)
    init_loss = - torch.nn.functional.cosine_similarity(embedding_layer(g.edge_features['idx']).sum(dim=-1), target,
                                                        dim=0)
    for i in range(100):
        g.edge_features['edge_emb'] = embedding_layer(g.edge_features['idx'])
        node_emb_sum = g.edge_features['edge_emb'].sum(dim=-1)
        loss = - torch.nn.functional.cosine_similarity(node_emb_sum, target, dim=0)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
    print('Initial loss = {}. Final loss = {}'.format(init_loss, loss))
    assert init_loss > loss


def test_conversion_dgl():
    g = GraphData()
    g.add_nodes(10)
    for i in range(10):
        g.add_edge(src=i, tgt=(i + 1) % 10)
    g.node_features['node_feat'] = torch.randn((10, 10))
    g.node_features['zero'] = torch.zeros(10)
    g.node_features['idx'] = torch.tensor(list(range(10)), dtype=torch.long)
    g.edge_features['edge_feat'] = torch.randn((10, 10))
    g.edge_features['idx'] = torch.tensor(list(range(10)), dtype=torch.long)
    # Test to_dgl
    dgl_g = g.to_dgl()
    for node_feat_name in g.get_node_feature_names():
        if g.node_features[node_feat_name] is None:
            assert node_feat_name not in dgl_g.ndata.keys()
        else:
            assert torch.all(torch.eq(dgl_g.ndata[node_feat_name], g.node_features[node_feat_name]))
    for edge_feat_name in g.get_edge_feature_names():
        if g.edge_features[edge_feat_name] is None:
            assert edge_feat_name not in dgl_g.edata.keys()
        else:
            assert torch.all(torch.eq(dgl_g.edata[edge_feat_name], g.edge_features[edge_feat_name]))
    assert g.get_node_num() == dgl_g.number_of_nodes()
    src, tgt = dgl_g.all_edges()
    dgl_g_edges = []
    for i in range(src.shape[0]):
        dgl_g_edges.append((int(src[i]), int(tgt[i])))
    assert g.get_all_edges() == dgl_g_edges
    # Test from_dgl
    g1 = from_dgl(dgl_g)
    for node_feat_name in g.get_node_feature_names():
        try:
            assert torch.all(torch.eq(g1.node_features[node_feat_name], g.node_features[node_feat_name]))
        except TypeError:
            assert g1.node_features[node_feat_name] == g.node_features[node_feat_name]
    for edge_feat_name in g.get_edge_feature_names():
        try:
            assert torch.all(torch.eq(g1.edge_features[edge_feat_name], g.edge_features[edge_feat_name]))
        except TypeError:
            assert g1.edge_features[edge_feat_name] == g.edge_features[edge_feat_name]
    assert g1.get_node_num() == g.get_node_num()
    assert g1.get_all_edges() == g.get_all_edges()


def test_batch():
    g_list = []
    batched_edges = []
    graph_edges_list = []
    for i in range(5):
        g = GraphData()
        g.add_nodes(10)
        for j in range(10):
            g.add_edge(src=j, tgt=(j + 1) % 10)
            batched_edges.append((i * 10 + j, i * 10 + ((j + 1) % 10)))
        g.node_features['idx'] = torch.ones(10) * i
        g.edge_features['idx'] = torch.ones(10) * i
        graph_edges_list.append(g.get_all_edges())
        g_list.append(g)

    # Test to_batch
    batch = to_batch(g_list)

    target_batch_idx = []
    for i in range(5):
        for j in range(10):
            target_batch_idx.append(i)

    assert batch.batch == target_batch_idx
    assert batch.get_node_num() == 50
    assert batch.get_all_edges() == batched_edges

    # Test from_batch
    graph_list = from_batch(batch)

    for i in range(len(graph_list)):
        g = graph_list[i]
        assert g.get_all_edges() == graph_edges_list[i]
        assert g.get_node_num() == 10
        assert torch.all(torch.eq(g.node_features['idx'], torch.ones(10) * i))
        assert torch.all(torch.eq(g.edge_features['idx'], torch.ones(10) * i))

    # Test graph with 0 edges
    gl = [GraphData() for _ in range(5)]
    for i in range(5):
        gl[i].add_nodes(1)
    b = to_batch(gl)
    gll = from_batch(b)
    for i in range(5):
        print(gll[i].get_edge_num())



import torch

from .coverage_loss import CoverageLoss
from .general_loss import GeneralLoss
from .base import GeneralLossBase
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab


class SeqGenerationLoss(GeneralLossBase):
    """
        The general loss for ``Graph2Seq`` model.
    Parameters
    ----------
    vocab: Vocab
        The vocab for loss calculation.
    use_coverage: bool, default=False
        Whether use coverage mechanism. If set ``True``, the we will add the coverage loss.
    coverage_weight: float, default=0.3
        The weight of coverage loss.
    """
    def __init__(self, ignore_index, use_coverage=False, coverage_weight=0.3):
        super(SeqGenerationLoss, self).__init__()
        self.use_coverage = use_coverage
        self.loss_ce = GeneralLoss(loss_type="NLL", size_average=True, reduce=True,
                                   ignore_index=ignore_index)
        self.loss_coverage = CoverageLoss(cover_loss=coverage_weight)

    def forward(self, logits, label, enc_attn_weights=None, coverage_vectors=None):
        """
            The calculation method.
        Parameters
        ----------
        logits: torch.Tensor
            The probability with the shape of ``[batch_size, max_decoder_step, vocab_size]``. \
            Note that it is calculated by ``softmax``.

        label: torch.Tensor
            The ground-truth with the shape of ``[batch_size, max_decoder_step]``.
        enc_attn_weights: list[torch.Tensor], default=None
            The list containing all decoding steps' attention weights.
            The length should be the decoding step.
            Each element should be the tensor.
        coverage_vectors: list[torch.Tensor], default=None
            The list containing all coverage vectors in decoding module.

        Returns
        -------
        graph2seq_loss: torch.Tensor
        """
        loss_ce = self.loss_ce(torch.log(logits + 1e-31).transpose(1, 2), label)
        if self.use_coverage:
            loss_cover = self.loss_coverage(enc_attn_weights, coverage_vectors)
            loss_ce += loss_cover
        return loss_ce


from .ComplEx import ComplEx
from .ComplExLayer import ComplExLayer
from .DistMult import DistMult
from .DistMultLayer import DistMultLayer
from .TransE import TransE
from .TransELayer import TransELayer

__all__ = ['ComplEx', 'ComplExLayer', 'DistMult', 'DistMultLayer',
            'TransE', 'TransELayer']


from torch import nn
import torch
from ..base import LinkPredictionBase
from .ElementSumLayer import ElementSumLayer

class ElementSum(LinkPredictionBase):
    r"""Specific class for link prediction task.

    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    num_class : int 
               The number of node catrgoriey for classification
    hidden_size : list of int type values
                  Example for two layers's FeedforwardNN: [50, 20]
    activation: the activation function class for each fully connected layer
                Default: nn.ReLU()
                Example: nn.ReLU(),nn.Sigmoid().        

    """     
    def __init__(self, input_size,  hidden_size,num_class,activation=nn.ReLU()):        
        super(ElementSum, self).__init__()
        
        
        self.classifier=ElementSumLayer(input_size, num_class, hidden_size,activation)

    def forward(self, input_graph):
        r"""
        Forward functions to compute the logits tensor for link prediction.
    
      
        Parameters
        ----------
    
        input graph : GraphData
                     The tensors stored in the node feature field named "node_emb"  in the 
                     input_graph are used  for link prediction.

    
        Returns 
        ---------
        
        output_graph : GraphData
                      The computed logit tensor for each pair of nodes in the graph are stored
                      in the node feature field named "edge_logits".
                      logit tensor shape is: [num_class] 
        """ 
        #get the nod embedding from the graph 
        node_emb=input_graph.node_features['node_emb']
        
        #add the edges and edge prediction logits into the graph
        num_node=node_emb.shape[0]
        node_idx_list=[idx for idx in range(num_node)]
        src_idx=torch.tensor(node_idx_list).view(-1,1).repeat(1,num_node).view(-1)
        dst_idx=torch.tensor(node_idx_list).view(1,-1).repeat(num_node,1).view(-1)
        input_graph.add_edges(src_idx,dst_idx)
        input_graph.edge_features['logits']=self.classifier(node_emb)      
         
        return input_graph







import pickle as pkl

import networkx as nx
import numpy as np
import scipy.sparse as sp
import torch
from sklearn.metrics import roc_auc_score, average_precision_score


def load_data(dataset):
    # load the data: x, tx, allx, graph
    names = ['x', 'tx', 'allx', 'graph']
    objects = []
    for i in range(len(names)):
        '''
        fix Pickle incompatibility of numpy arrays between Python 2 and 3
        https://stackoverflow.com/questions/11305790/pickle-incompatibility-of-numpy-arrays-between-python-2-and-3
        '''
        with open("data/ind.{}.{}".format(dataset, names[i]), 'rb') as rf:
            u = pkl._Unpickler(rf)
            u.encoding = 'latin1'
            cur_data = u.load()
            objects.append(cur_data)
        # objects.append(
        #     pkl.load(open("data/ind.{}.{}".format(dataset, names[i]), 'rb')))
    x, tx, allx, graph = tuple(objects)
    test_idx_reorder = parse_index_file(
        "data/ind.{}.test.index".format(dataset))
    test_idx_range = np.sort(test_idx_reorder)

    if dataset == 'citeseer':
        # Fix citeseer dataset (there are some isolated nodes in the graph)
        # Find isolated nodes, add them as zero-vecs into the right position
        test_idx_range_full = range(
            min(test_idx_reorder), max(test_idx_reorder) + 1)
        tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
        tx_extended[test_idx_range - min(test_idx_range), :] = tx
        tx = tx_extended

    features = sp.vstack((allx, tx)).tolil()
    features[test_idx_reorder, :] = features[test_idx_range, :]
    features = torch.FloatTensor(np.array(features.todense()))
    adj = nx.adjacency_matrix(nx.from_dict_of_lists(graph))

    return adj, features


def parse_index_file(filename):
    index = []
    for line in open(filename):
        index.append(int(line.strip()))
    return index


def sparse_to_tuple(sparse_mx):
    if not sp.isspmatrix_coo(sparse_mx):
        sparse_mx = sparse_mx.tocoo()
    coords = np.vstack((sparse_mx.row, sparse_mx.col)).transpose()
    values = sparse_mx.data
    shape = sparse_mx.shape
    return coords, values, shape


def mask_test_edges(adj):
    # Function to build test set with 10% positive links
    # NOTE: Splits are randomized and results might slightly deviate from reported numbers in the paper.
    # TODO: Clean up.

    # Remove diagonal elements
    adj = adj - sp.dia_matrix((adj.diagonal()[np.newaxis, :], [0]), shape=adj.shape)
    adj.eliminate_zeros()
    # Check that diag is zero:
    assert np.diag(adj.todense()).sum() == 0

    adj_triu = sp.triu(adj)
    adj_tuple = sparse_to_tuple(adj_triu)
    edges = adj_tuple[0]
    edges_all = sparse_to_tuple(adj)[0]
    num_test = int(np.floor(edges.shape[0] / 10.))
    num_val = int(np.floor(edges.shape[0] / 20.))

    all_edge_idx = list(range(edges.shape[0]))
    np.random.shuffle(all_edge_idx)
    val_edge_idx = all_edge_idx[:num_val]
    test_edge_idx = all_edge_idx[num_val:(num_val + num_test)]
    test_edges = edges[test_edge_idx]
    val_edges = edges[val_edge_idx]
    train_edges = np.delete(edges, np.hstack([test_edge_idx, val_edge_idx]), axis=0)

    def ismember(a, b, tol=5):
        rows_close = np.all(np.round(a - b[:, None], tol) == 0, axis=-1)
        return np.any(rows_close)

    test_edges_false = []
    while len(test_edges_false) < len(test_edges):
        idx_i = np.random.randint(0, adj.shape[0])
        idx_j = np.random.randint(0, adj.shape[0])
        if idx_i == idx_j:
            continue
        if ismember([idx_i, idx_j], edges_all):
            continue
        if test_edges_false:
            if ismember([idx_j, idx_i], np.array(test_edges_false)):
                continue
            if ismember([idx_i, idx_j], np.array(test_edges_false)):
                continue
        test_edges_false.append([idx_i, idx_j])

    val_edges_false = []
    while len(val_edges_false) < len(val_edges):
        idx_i = np.random.randint(0, adj.shape[0])
        idx_j = np.random.randint(0, adj.shape[0])
        if idx_i == idx_j:
            continue
        if ismember([idx_i, idx_j], train_edges):
            continue
        if ismember([idx_j, idx_i], train_edges):
            continue
        if ismember([idx_i, idx_j], val_edges):
            continue
        if ismember([idx_j, idx_i], val_edges):
            continue
        if val_edges_false:
            if ismember([idx_j, idx_i], np.array(val_edges_false)):
                continue
            if ismember([idx_i, idx_j], np.array(val_edges_false)):
                continue
        val_edges_false.append([idx_i, idx_j])

    assert ~ismember(test_edges_false, edges_all)
    assert ~ismember(val_edges_false, edges_all)
    assert ~ismember(val_edges, train_edges)
    assert ~ismember(test_edges, train_edges)
    assert ~ismember(val_edges, test_edges)

    data = np.ones(train_edges.shape[0])

    # Re-build adj matrix
    adj_train = sp.csr_matrix((data, (train_edges[:, 0], train_edges[:, 1])), shape=adj.shape)
    adj_train = adj_train + adj_train.T

    # NOTE: these edge lists only contain single direction of edge!
    return adj_train, train_edges, val_edges, val_edges_false, test_edges, test_edges_false


def preprocess_graph(adj):
    adj = sp.coo_matrix(adj)
    adj_ = adj + sp.eye(adj.shape[0])
    rowsum = np.array(adj_.sum(1))
    degree_mat_inv_sqrt = sp.diags(np.power(rowsum, -0.5).flatten())
    adj_normalized = adj_.dot(degree_mat_inv_sqrt).transpose().dot(degree_mat_inv_sqrt).tocoo()
    # return sparse_to_tuple(adj_normalized)
    return sparse_mx_to_torch_sparse_tensor(adj_normalized)


def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)


# def get_roc_score(adj_rec,adj_orig, edges_pos, edges_neg):
#     preds = []
#     pos = []
#     for e in edges_pos:
#         preds.append(adj_rec[e[0], e[1]]) #M*2
#         pos.append(adj_orig[e[0], e[1]]) #M

#     preds_neg = []
#     neg = []
#     for e in edges_neg:
#         preds_neg.append(adj_rec[e[0], e[1]])
#         neg.append(adj_orig[e[0], e[1]])

#     preds_all = np.hstack([preds, preds_neg])
#     labels_all = np.hstack([np.ones(len(preds)), np.zeros(len(preds_neg))])
#     roc_score = roc_auc_score(labels_all, preds_all)
#     ap_score = average_precision_score(labels_all, preds_all)

#     return roc_score, ap_score


def get_roc_score(adj_rec, adj_orig, edges_pos, edges_neg):
    def sigmoid(x):
        return 1 / (1 + np.exp(-x))
    # Predict on test set of edges

    preds = []
    pos = []
    for e in edges_pos:
        preds.append(sigmoid(adj_rec[e[0], e[1]]))
        pos.append(adj_orig[e[0], e[1]])
    preds_neg = []
    neg = []
    for e in edges_neg:
        preds_neg.append(sigmoid(adj_rec[e[0], e[1]]))
        neg.append(adj_orig[e[0], e[1]])
    preds_all = np.hstack([preds, preds_neg])
    labels_all = np.hstack([np.ones(len(preds)), np.zeros(len(preds_neg))])
    roc_score = roc_auc_score(labels_all, preds_all)
    ap_score = average_precision_score(labels_all, preds_all)
    return roc_score, ap_score


import os

os.environ["CUDA_VISIBLE_DEVICES"] = "2"

from .dataset import CNNDataset
from .model_g2s import Graph2seq
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel
from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals_no_size

import numpy as np
import torch
from torch.utils.data import DataLoader
import torch.optim as optim
from .config_g2s import get_args
from .utils import get_log, wordid2str
from graph4nlp.pytorch.modules.evaluation.rouge import ROUGE
import time
from graph4nlp.pytorch.data.data import from_batch, GraphData
import copy
from graph4nlp.pytorch.modules.utils.padding_utils import pad_2d_vals_no_size
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab

class CNN:
    def __init__(self, opt):
        super(CNN, self).__init__()
        self.opt = opt
        self._build_device(self.opt)
        self._build_logger(self.opt.log_file)
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_device(self, opt):
        seed = opt.seed
        np.random.seed(seed)
        if opt.use_gpu != 0 and torch.cuda.is_available():
            print('[ Using CUDA ]')
            torch.manual_seed(seed)
            torch.cuda.manual_seed_all(seed)
            from torch.backends import cudnn
            cudnn.benchmark = True
            device = torch.device('cuda' if opt.gpu < 0 else 'cuda:%d' % opt.gpu)
        else:
            print('[ Using CPU ]')
            device = torch.device('cpu')
        self.device = device

    def _build_logger(self, log_file):
        self.logger = get_log(log_file)

    def _build_dataloader(self):
        if 'DependencyGraph' in self.opt.topology_subdir:
            graph_type = 'static'
            topology_builder = DependencyBasedGraphConstruction
            topology_subdir = self.opt.topology_subdir
            dynamic_graph_type = None
            dynamic_init_topology_builder = None
        elif self.opt.topology_subdir == 'node_emb':
            graph_type = 'dynamic'
            topology_builder = NodeEmbeddingBasedGraphConstruction
            topology_subdir = 'NodeEmb'
            dynamic_graph_type = 'node_emb'
            dynamic_init_topology_builder = DependencyBasedGraphConstruction
        else:
            raise NotImplementedError()

        dataset = CNNDataset(root_dir=self.opt.root_dir,
                             tokenizer=None,
                             device=self.device,
                             word_emb_size=self.opt.word_emb_size,
                             thread_number=35,
                             share_vocab=True,
                             graph_type=graph_type,
                             topology_builder=topology_builder,
                             topology_subdir=topology_subdir,
                             dynamic_graph_type=dynamic_graph_type,
                             dynamic_init_topology_builder=dynamic_init_topology_builder,
                             dynamic_init_topology_aux_args={'dummy_param': 0})

        self.train_dataloader = DataLoader(dataset.train, batch_size=self.opt.batch_size, shuffle=True, num_workers=0,
                                           collate_fn=dataset.collate_fn)
        self.val_dataloader = DataLoader(dataset.val, batch_size=self.opt.batch_size, shuffle=False, num_workers=0,
                                         collate_fn=dataset.collate_fn)
        self.test_dataloader = DataLoader(dataset.test, batch_size=self.opt.batch_size, shuffle=False, num_workers=0,
                                          collate_fn=dataset.collate_fn)
        self.vocab: VocabModel = dataset.vocab_model

    def _build_model(self):
        self.model = Graph2seq(self.vocab,
                               use_copy=self.opt.use_copy,
                               use_coverage=self.opt.use_coverage,
                               gnn=self.opt.gnn,
                               device=self.device,
                               rnn_dropout=self.opt.rnn_dropout,
                               emb_dropout=self.opt.word_dropout,
                               hidden_size=self.opt.hidden_size).to(self.device)

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=self.opt.learning_rate)

    def _build_evaluation(self):
        self.metrics = [ROUGE()]

    def train(self):
        max_score = -1
        self._best_epoch = -1
        for epoch in range(200):
            # score = self.evaluate(split="val")
            self.model.train()
            self.train_epoch(epoch, split="train")
            self._adjust_lr(epoch)
            if epoch >= 0:
                score = self.evaluate(split="val")
                if score >= max_score:
                    self.logger.info("Best model saved, epoch {}".format(epoch))
                    self.save_checkpoint("best.pth")
                    self._best_epoch = epoch
                max_score = max(max_score, score)
            if epoch >= 30 and self._stop_condition(epoch):
                break
        return max_score

    def _stop_condition(self, epoch, patience=2000):
        return epoch > patience + self._best_epoch

    def _adjust_lr(self, epoch):
        def set_lr(optimizer, decay_factor):
            for group in optimizer.param_groups:
                group['lr'] = group['lr'] * decay_factor

        epoch_diff = epoch - self.opt.lr_start_decay_epoch
        if epoch_diff >= 0 and epoch_diff % self.opt.lr_decay_per_epoch == 0:
            if self.opt.learning_rate > self.opt.min_lr:
                set_lr(self.optimizer, self.opt.lr_decay_rate)
                self.opt.learning_rate = self.opt.learning_rate * self.opt.lr_decay_rate
                self.logger.info("Learning rate adjusted: {:.5f}".format(self.opt.learning_rate))

    def prepare_ext_vocab(self, batch, vocab, gt_str=None):
        oov_dict = copy.deepcopy(vocab.in_word_vocab)
        for g in batch:
            token_matrix = []
            for node_idx in range(g.get_node_num()):
                node_token = g.node_attributes[node_idx]['token']
                if oov_dict.getIndex(node_token) == oov_dict.UNK:
                    oov_dict._add_words([node_token])
                token_matrix.append([oov_dict.getIndex(node_token)])
            token_matrix = torch.tensor(token_matrix, dtype=torch.long).to(self.device)
            g.node_features['token_id_oov'] = token_matrix

        if gt_str is not None:
            oov_tgt_collect = []
            for s in gt_str:
                oov_tgt = oov_dict.to_index_sequence(s)
                oov_tgt.append(oov_dict.EOS)
                oov_tgt = np.array(oov_tgt)
                oov_tgt_collect.append(oov_tgt)

            output_pad = pad_2d_vals_no_size(oov_tgt_collect)

            tgt_seq = torch.from_numpy(output_pad).long().to(self.device)
            return oov_dict, tgt_seq
        else:
            return oov_dict

    def train_epoch(self, epoch, split="train"):
        assert split in ["train"]
        self.logger.info("Start training in split {}, Epoch: {}".format(split, epoch))
        loss_collect = []
        dataloader = self.train_dataloader
        step_all_train = len(dataloader)
        start = time.time()
        for step, data in enumerate(dataloader):
            graph_list, tgt, gt_str = data
            tgt = tgt.to(self.device)
            oov_dict = None
            if self.opt.use_copy:
                oov_dict, tgt = self.prepare_ext_vocab(graph_list, self.vocab, gt_str=gt_str)

            _, loss = self.model(graph_list, tgt, oov_dict=oov_dict, require_loss=True)
            loss_collect.append(loss.item())
            if step % self.opt.loss_display_step == 0 and step != 0:
                self.logger.info("Epoch {}: [{} / {}] loss: {:.3f}".format(epoch, step, step_all_train,
                                                                           np.mean(loss_collect)))
                loss_collect = []
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

    @torch.no_grad()
    def evaluate(self, split="val", test_mode=False):
        self.model.eval()
        pred_collect = []
        gt_collect = []
        assert split in ["val", "test"]
        dataloader = self.val_dataloader if split == "val" else self.test_dataloader
        for data in dataloader:
            graph_list, tgt, gt_str = data
            if self.opt.use_copy:
                oov_dict = self.prepare_ext_vocab(graph_list, self.vocab)
                ref_dict = oov_dict
            else:
                oov_dict = None
                ref_dict = self.vocab.out_word_vocab

            prob = self.model(graph_list, oov_dict=oov_dict, require_loss=False)
            pred = prob.argmax(dim=-1)

            pred_str = wordid2str(pred.detach().cpu(), ref_dict)
            pred_collect.extend(pred_str)
            gt_collect.extend(gt_str)

        if test_mode==True:
            with open(self.opt.checkpoint_save_path+'/cnn_pred_output.txt','w+') as f:
                for line in pred_collect:
                    f.write(line+'\n')

            with open(self.opt.checkpoint_save_path+'/cnn_tgt_output.txt','w+') as f:
                for line in gt_collect:
                    f.write(line+'\n')

        score, _ = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
        self.logger.info("Evaluation ROUGE in `{}` split: {:.3f}".format(split, score))
        return score

    def load_checkpoint(self, checkpoint_name):
        checkpoint_path = os.path.join(self.opt.checkpoint_save_path, checkpoint_name)
        self.model.load_state_dict(torch.load(checkpoint_path))

    def save_checkpoint(self, checkpoint_name):
        checkpoint_path = os.path.join(self.opt.checkpoint_save_path, checkpoint_name)
        torch.save(self.model.state_dict(), checkpoint_path)


if __name__ == "__main__":
    opt = get_args()
    runner = CNN(opt)
    max_score = runner.train()
    print("Train finish, best val score: {:.3f}".format(max_score))
    runner.load_checkpoint('best.pth')
    runner.evaluate(split="test", test_mode=True)


"""Graph classificatioin example
Reference: https://github.com/snap-stanford/ogb/blob/master/examples/graphproppred/mol/main_pyg.py
"""
import time
import argparse
from tqdm import tqdm
import numpy as np

import torch
from torch import nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import DataLoader
from ogb.graphproppred import DglGraphPropPredDataset, Evaluator
# from dgl.data import register_data_args

from ...modules.graph_embedding.gat import GAT
from ...modules.prediction.classification.graph_classification import AvgPooling, FeedForwardNN


cls_criterion = torch.nn.BCEWithLogitsLoss()
reg_criterion = torch.nn.MSELoss()



class GNNClassifier(nn.Module):
    def __init__(self,
                num_layers,
                input_size,
                hidden_size,
                output_size,
                num_heads,
                num_out_heads,
                direction_option,
                graph_pooling,
                feat_drop=0.6,
                attn_drop=0.6,
                negative_slope=0.2,
                residual=False,
                activation=F.elu):
        super(GNNClassifier, self).__init__()
        self.direction_option = direction_option
        heads = [num_heads] * (num_layers - 1) + [num_out_heads]
        self.gnn = GAT(num_layers,
                    input_size,
                    hidden_size,
                    hidden_size,
                    heads,
                    direction_option=direction_option,
                    feat_drop=feat_drop,
                    attn_drop=attn_drop,
                    negative_slope=negative_slope,
                    residual=residual,
                    activation=activation)

        self.clf = FeedForwardNN(2 * hidden_size if self.direction_option == 'bi_sep' else hidden_size,
                        output_size,
                        [hidden_size],
                        graph_pool_type=graph_pooling)


    def forward(self, batched_graph):
        batched_graph = self.gnn(batched_graph)
        batched_graph = self.clf(batched_graph)
        logits = batched_graph.graph_attributes['logits']

        return logits


def train(model, device, loader, optimizer, task_type):
    model.train()

    for step, batch in enumerate(tqdm(loader, desc="Iteration")):
        batch = batch.to(device)

        if batch.x.shape[0] == 1 or batch.batch[-1] == 0:
            pass
        else:
            pred = model(batch)
            optimizer.zero_grad()
            ## ignore nan targets (unlabeled) when computing training loss.
            is_labeled = batch.y == batch.y
            if "classification" in task_type:
                loss = cls_criterion(pred.to(torch.float32)[is_labeled], batch.y.to(torch.float32)[is_labeled])
            else:
                loss = reg_criterion(pred.to(torch.float32)[is_labeled], batch.y.to(torch.float32)[is_labeled])
            loss.backward()
            optimizer.step()

def eval(model, device, loader, evaluator):
    model.eval()
    y_true = []
    y_pred = []

    for step, batch in enumerate(tqdm(loader, desc="Iteration")):
        batch = batch.to(device)

        if batch.x.shape[0] == 1:
            pass
        else:
            with torch.no_grad():
                pred = model(batch)

            y_true.append(batch.y.view(pred.shape).detach().cpu())
            y_pred.append(pred.detach().cpu())

    y_true = torch.cat(y_true, dim = 0).numpy()
    y_pred = torch.cat(y_pred, dim = 0).numpy()

    input_dict = {"y_true": y_true, "y_pred": y_pred}

    return evaluator.eval(input_dict)




def main():
    # Training settings
    parser = argparse.ArgumentParser(description='GAT')
    # register_data_args(parser)
    parser.add_argument("--graph-pooling", type=str, default='avg_pool',
                        help="graph pooling (`avg_pool`, `max_pool`)")
    parser.add_argument("--direction-option", type=str, default='uni',
                        help="direction type (`uni`, `bi_fuse`, `bi_sep`)")
    parser.add_argument("--num-heads", type=int, default=8,
                        help="number of hidden attention heads")
    parser.add_argument("--num-out-heads", type=int, default=1,
                        help="number of output attention heads")
    parser.add_argument("--num-layers", type=int, default=2,
                        help="number of hidden layers")
    parser.add_argument("--num-hidden", type=int, default=8,
                        help="number of hidden units")
    parser.add_argument("--residual", action="store_true", default=False,
                        help="use residual connection")
    parser.add_argument("--in-drop", type=float, default=.6,
                        help="input feature dropout")
    parser.add_argument("--attn-drop", type=float, default=.6,
                        help="attention dropout")
    parser.add_argument("--lr", type=float, default=0.005,
                        help="learning rate")
    parser.add_argument('--weight-decay', type=float, default=5e-4,
                        help="weight decay")
    parser.add_argument('--negative-slope', type=float, default=0.2,
                        help="the negative slope of leaky relu")
    parser.add_argument('--early-stop', action='store_true', default=False,
                        help="indicates whether to use early stop or not")
    parser.add_argument("--patience", type=int, default=100,
                        help="early stopping patience")


    parser.add_argument('--device', type=int, default=0,
                        help='which gpu to use if any (default: 0)')
    parser.add_argument('--drop_ratio', type=float, default=0.5,
                        help='dropout ratio (default: 0.5)')
    parser.add_argument('--batch_size', type=int, default=32,
                        help='input batch size for training (default: 32)')
    parser.add_argument('--epochs', type=int, default=100,
                        help='number of epochs to train (default: 100)')
    parser.add_argument('--num_workers', type=int, default=0,
                        help='number of workers (default: 0)')
    parser.add_argument('--dataset', type=str, default="ogbg-molhiv",
                        help='dataset name (default: ogbg-molhiv)')
    parser.add_argument('--feature', type=str, default="full",
                        help='full feature or simple feature')
    parser.add_argument('--filename', type=str, default="",
                        help='filename to output result (default: )')
    args = parser.parse_args()

    device = torch.device("cuda:" + str(args.device)) if torch.cuda.is_available() else torch.device("cpu")

    ### automatic dataloading and splitting
    dataset = DglGraphPropPredDataset(name = args.dataset)

    if args.feature == 'full':
        pass
    elif args.feature == 'simple':
        print('using simple feature')
        # only retain the top two node/edge features
        dataset.data.x = dataset.data.x[:,:2]
        dataset.data.edge_attr = dataset.data.edge_attr[:,:2]

    split_idx = dataset.get_idx_split()

    ### automatic evaluator. takes dataset name as input
    evaluator = Evaluator(args.dataset)

    train_loader = DataLoader(dataset[split_idx["train"]], batch_size=args.batch_size, shuffle=True, num_workers = args.num_workers)
    valid_loader = DataLoader(dataset[split_idx["valid"]], batch_size=args.batch_size, shuffle=False, num_workers = args.num_workers)
    test_loader = DataLoader(dataset[split_idx["test"]], batch_size=args.batch_size, shuffle=False, num_workers = args.num_workers)

    num_feats = dataset.graphs[0].ndata['feat'].shape
    n_classes = int(dataset.num_classes)
    model = GNNClassifier(args.num_layers,
                num_feats,
                args.num_hidden,
                n_classes,
                args.num_heads,
                args.num_out_heads,
                direction_option=args.direction_option,
                graph_pooling=args.graph_pooling,
                feat_drop=args.in_drop,
                attn_drop=args.attn_drop,
                negative_slope=args.negative_slope,
                residual=args.residual,
                activation=F.elu)

    optimizer = optim.Adam(model.parameters(), lr=args.lr)

    valid_curve = []
    test_curve = []
    train_curve = []

    for epoch in range(1, args.epochs + 1):
        print("=====Epoch {}".format(epoch))
        print('Training...')
        train(model, device, train_loader, optimizer, dataset.task_type)

        print('Evaluating...')
        train_perf = eval(model, device, train_loader, evaluator)
        valid_perf = eval(model, device, valid_loader, evaluator)
        test_perf = eval(model, device, test_loader, evaluator)

        print({'Train': train_perf, 'Validation': valid_perf, 'Test': test_perf})

        train_curve.append(train_perf[dataset.eval_metric])
        valid_curve.append(valid_perf[dataset.eval_metric])
        test_curve.append(test_perf[dataset.eval_metric])

    if 'classification' in dataset.task_type:
        best_val_epoch = np.argmax(np.array(valid_curve))
        best_train = max(train_curve)
    else:
        best_val_epoch = np.argmin(np.array(valid_curve))
        best_train = min(train_curve)

    print('Finished training!')
    print('Best validation score: {}'.format(valid_curve[best_val_epoch]))
    print('Test score: {}'.format(test_curve[best_val_epoch]))

    if not args.filename == '':
        torch.save({'Val': valid_curve[best_val_epoch], 'Test': test_curve[best_val_epoch], 'Train': train_curve[best_val_epoch], 'BestTrain': best_train}, args.filename)


if __name__ == "__main__":
    main()



from graph4nlp.pytorch.modules.graph_construction import IEBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.utils.vocab_utils import VocabModel

embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
                           'seq_info_encode_strategy': "bilstm"}
vocab = VocabModel.build("/home/shiina/shiina/lib/graph4nlp/graph4nlp/pytorch/test/dataset/graph2seq/processed/DependencyGraph/vocab.pt")
graph_topology = IEBasedGraphConstruction(embedding_style=embedding_style, vocab=vocab.in_word_vocab)
dep = DependencyBasedGraphConstruction(embedding_style=embedding_style, vocab=vocab.in_word_vocab)
print(isinstance(graph_topology, IEBasedGraphConstruction))
print(type(graph_topology) == IEBasedGraphConstruction)




import random
import numpy as np
import torch
from stanfordcorenlp import StanfordCoreNLP

from ...modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from ...modules.graph_construction.embedding_construction import EmbeddingConstruction
from ...modules.utils.vocab_utils import VocabModel
from ...data.data import GraphData


if __name__ == "__main__":
    seed = 1234
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)

    # some tests for data structure code.
    
    # g = GraphData()
    # g.add_nodes(1)
    # g.add_nodes(1)
    # print(g.node_attributes[1]['node_attr'])
    # print(g.get_node_attrs(1))
    # g.node_attributes[1]['word'] = "word"
    # print(g.get_node_attrs(1))

    # print(g.node_features)
    # g.node_features['node_feat'] = torch.zeros(g.get_node_num(), 128)
    # print(g.node_features)
    # g.set_node_features(1, {"node_feat":torch.ones(1, 128)})
    # print(g.get_node_features(1))

    # g.add_edge(0,1)
    # g.add_edge(0,1)
    # print(g.get_edge_num())
    # print(g.edges)
    # print(list(g.nodes))
    # raise BaseException

    raw_data = "I love you. My motherland."
    # raw_data = "James went to the corner-shop. He want to buy some (eggs), <milk> and bread for breakfast."
    vocab_model = VocabModel(raw_data, max_word_vocab_size=None,
                                min_word_vocab_freq=1,
                                word_emb_size=300)

    embedding_styles = {
        'word_emb_type': 'w2v',
        'node_edge_level_emb_type': 'mean',
        'graph_level_emb_type': 'identity',
    }

    nlp_parser = StanfordCoreNLP('http://localhost', port=9000, timeout=300000)
    print("syntactic parser ready\n-------------------")

    # constituency_graph_gonstructor = ConstituencyBasedGraphConstruction(hidden_emb_size=128, embedding_style=embedding_styles, word_emb_size=300, vocab=vocab_model.word_vocab)
    # for sentence in raw_data:
        # output_graph = constituency_graph_gonstructor.forward(sentence[0], nlp_parser)
    output_graph = ConstituencyBasedGraphConstruction.topology(raw_data, nlp_parser)
    print(output_graph.node_attributes)
    print(output_graph.edges)
    for _edge in output_graph.get_all_edges():
        print(output_graph.nodes[_edge[0]].attributes['token'], output_graph.nodes[_edge[1]].attributes['token'])
    # print("-----------------------\nvocab size")
    # print(vocab_model.word_vocab.word2index)


from ..base import KGCompletionBase
from .ComplExLayer import ComplExLayer
from .....data.data import GraphData


class ComplEx(KGCompletionBase):
    r"""Specific class for knowledge graph completion task.

    ComplEx from paper `Complex Embeddings for Simple Link Prediction
    <http://proceedings.mlr.press/v48/trouillon16.pdf>`__.


    Parameters
    ----------
    input_dropout: float
        Dropout for node_emb and rel_emb. Default: 0.0

    rel_emb_from_gnn: bool
        If `rel_emb` is computed from GNN, rel_emb_from_gnn is set to `True`.
        Else, rel_emb is initialized as nn.Embedding randomly. Default: `True`.

    num_relations: int
        Number of relations. `num_relations` is needed if rel_emb_from_gnn==True.
        Default: `None`.

    embedding_dim: int
        Dimension of the rel_emb. `embedding_dim` is needed if rel_emb_from_gnn==True.
        Default: `0`.

    edge2node: bool
        The edges or relations in KG are converted to nodes. Default: `False`.

    loss_name: str
        The loss type selected fot the KG completion task. Default: `'BCELoss'`
    """

    def __init__(self,
                 input_dropout=0.0,
                 rel_emb_from_gnn=True,
                 num_relations=None,
                 embedding_dim=None,
                 edge2node=False,
                 loss_name='BCELoss'):
        super(ComplEx, self).__init__()
        self.rel_emb_from_gnn = rel_emb_from_gnn
        self.edge2node = edge2node
        self.loss_name = loss_name
        self.classifier = ComplExLayer(input_dropout, rel_emb_from_gnn,
                                       num_relations, embedding_dim, loss_name)


    def forward(self, input_graph: GraphData):
        r"""
        Forward functions to compute the logits tensor for kg completion.

        Parameters
        ----------

        input graph : GraphData
                     The tensors stored in the node feature field named "node_emb" and
                     "rel_emb" in the input_graph are used for knowledge graph completion.

        Returns
        ---------

        output_graph : GraphData
                      The computed logit tensor for each nodes in the graph are stored
                      in the node feature field named "node_logits".
                      logit tensor shape is: [num_class]
        """

        node_emb_real = input_graph.node_features['node_emb_real']
        node_emb_img = input_graph.node_features['node_emb_img']
        if self.loss_name in ['SoftplusLoss', 'SigmoidLoss']:
            multi_label = input_graph.graph_attributes['multi_binary_label']  # [B, N]
        else:
            multi_label = None

        if self.edge2node:
            rel_emb_real = node_emb_real
            rel_emb_img = node_emb_img
        else:
            if 'rel_emb_real' in input_graph.edge_features.keys() and \
                    'rel_emb_img' in input_graph.edge_features.keys():
                rel_emb_real = input_graph.edge_features['rel_emb_real']
                rel_emb_img = input_graph.edge_features['rel_emb_img']
            else:
                assert self.rel_emb_from_gnn == False
                rel_emb_real = None
                rel_emb_img = None

        if 'list_e_r_pair_idx' in input_graph.graph_attributes.keys():
            list_e_r_pair_idx = input_graph.graph_attributes['list_e_r_pair_idx']
            list_e_e_pair_idx = None
        elif 'list_e_e_pair_idx' in input_graph.graph_attributes.keys():
            list_e_e_pair_idx = input_graph.graph_attributes['list_e_e_pair_idx']
            list_e_r_pair_idx = None
        else:
            raise RuntimeError("'list_e_r_pair_idx' or 'list_e_e_pair_idx' should be given.")

        if multi_label is None:
            input_graph.graph_attributes['logits'] = self.classifier(node_emb_real,
                                                                  node_emb_img,
                                                                  rel_emb_real,
                                                                  rel_emb_img,
                                                                  list_e_r_pair_idx,
                                                                  list_e_e_pair_idx)
        else:
            input_graph.graph_attributes['logits'], input_graph.graph_attributes['p_score'], \
            input_graph.graph_attributes['n_score'] = self.classifier(node_emb_real,
                                                                   node_emb_img,
                                                                   rel_emb_real,
                                                                   rel_emb_img,
                                                                   list_e_r_pair_idx,
                                                                   list_e_e_pair_idx,
                                                                   multi_label)

        return input_graph


import os
import random
import time
import pickle
import argparse
import copy

import dgl
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init
import torch.optim as optim
from stanfordcorenlp import StanfordCoreNLP

from graph4nlp.pytorch.data.data import GraphData, from_batch

from graph4nlp.pytorch.datasets.jobs import JobsDatasetForTree

from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.constituency_graph_construction import ConstituencyBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_graph_construction import NodeEmbeddingBasedGraphConstruction
from graph4nlp.pytorch.modules.graph_construction.node_embedding_based_refined_graph_construction import NodeEmbeddingBasedRefinedGraphConstruction

from graph4nlp.pytorch.modules.utils.tree_utils import to_cuda
from graph4nlp.pytorch.models.graph2tree import Graph2Tree
from graph4nlp.pytorch.modules.prediction.generation.TreeBasedDecoder import StdTreeDecoder, create_mask
from graph4nlp.pytorch.modules.utils.tree_utils import DataLoaderForGraphEncoder, Tree, Vocab, VocabForAll, to_cuda

import warnings

warnings.filterwarnings('ignore')

class Jobs:
    def __init__(self, opt=None):
        super(Jobs, self).__init__()
        self.opt = opt
        
        seed = opt["seed"]
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True

        if opt["gpuid"] == -1:
            self.device = torch.device("cpu")
        else:
            self.device = torch.device("cuda:{}".format(opt["gpuid"]))
        self.use_copy = True if self.opt["decoder_args"]["rnn_decoder_share"]["use_copy"] == 1 else False
        self.revectorization = True
        self.data_dir = opt["data_dir"]
        self.checkpoint_dir = opt["checkpoint_dir"]

        self._build_dataloader()
        self._build_model()
        self._build_optimizer()

    def _build_dataloader(self):
        use_copy = self.use_copy
        use_share_vocab = True

        if self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "dependency":
            topology_builder = DependencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "constituency":
            topology_builder = ConstituencyBasedGraphConstruction
            graph_type = 'static'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "node_emb":
            topology_builder = NodeEmbeddingBasedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_topology_builder = None
        elif self.opt["graph_construction_args"]["graph_construction_share"]["graph_type"] == "node_emb_refined":
            topology_builder = NodeEmbeddingBasedRefinedGraphConstruction
            graph_type = 'dynamic'
            dynamic_init_graph_type = self.opt["graph_construction_args"]["graph_construction_private"][
                "dynamic_init_graph_type"]
            if dynamic_init_graph_type is None or dynamic_init_graph_type == 'line':
                dynamic_init_topology_builder = None
            elif dynamic_init_graph_type == 'dependency':
                dynamic_init_topology_builder = DependencyBasedGraphConstruction
            elif dynamic_init_graph_type == 'constituency':
                dynamic_init_topology_builder = ConstituencyBasedGraphConstruction
            else:
                # dynamic_init_topology_builder
                raise RuntimeError('Define your own dynamic_init_topology_builder')
        else:
            raise NotImplementedError("Define your topology builder.")

        dataset = JobsDatasetForTree(root_dir=self.data_dir,
                                         seed=self.opt["seed"],
                                         word_emb_size=self.opt["graph_embedding_args"]["graph_embedding_share"]["input_size"],
                                         edge_strategy=self.opt["graph_construction_args"]["graph_construction_private"][
                                            "edge_strategy"],
                                         topology_builder=topology_builder,
                                         topology_subdir=self.opt["graph_construction_args"]["graph_construction_share"][
                                            "topology_subdir"], 
                                         graph_type=graph_type,
                                         dynamic_graph_type=self.opt["graph_construction_args"]["graph_construction_share"][
                                            "graph_type"],
                                         share_vocab=use_share_vocab,
                                         enc_emb_size=self.opt["graph_embedding_args"]["graph_embedding_share"]["input_size"],
                                         dec_emb_size=self.opt["decoder_args"]["rnn_decoder_share"]["input_size"],
                                         dynamic_init_topology_builder=dynamic_init_topology_builder, 
                                         device=self.device,
                                         min_freq=self.opt["min_freq"])

        self.train_data_loader = DataLoaderForGraphEncoder(
            use_copy=use_copy, use_share_vocab=use_share_vocab, data=dataset.train, dataset=dataset, mode="train", batch_size=self.opt["batch_size"], device=self.device)
        print("train sample size:", len(self.train_data_loader.data))
        self.test_data_loader = DataLoaderForGraphEncoder(
            use_copy=use_copy, use_share_vocab=use_share_vocab, data=dataset.test, dataset=dataset, mode="test", batch_size=1, device=self.device)
        print("test sample size:", len(self.test_data_loader.data))

        self.src_vocab = self.train_data_loader.src_vocab
        self.tgt_vocab = self.train_data_loader.tgt_vocab

        if use_share_vocab:
            self.share_vocab = self.train_data_loader.share_vocab
        print(self.share_vocab.symbol2idx)
        print("---Loading data done---\n")

    def _build_model(self):
        self.model = Graph2Tree.from_args(self.opt, VocabForAll(self.src_vocab, self.tgt_vocab, self.share_vocab), self.device)
        self.model.init(self.opt["init_weight"])
        self.model = to_cuda(self.model, self.device)
        print(self.model)

    def _build_optimizer(self):
        optim_state = {"learningRate": self.opt["learning_rate"],
                       "weight_decay": self.opt["weight_decay"]}
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(
            parameters, lr=optim_state['learningRate'], weight_decay=optim_state['weight_decay'])

    def prepare_ext_vocab(self, batch_graphs, src_vocab):
        oov_dict = copy.deepcopy(src_vocab)
        for g in batch_graphs:
            token_matrix = []
            for node_idx in range(g.get_node_num()):
                node_token = g.node_attributes[node_idx]['token']
                if (g.node_attributes[node_idx].get('type') == None or g.node_attributes[node_idx].get('type') == 0) \
                        and oov_dict.get_symbol_idx(node_token) == oov_dict.get_symbol_idx(oov_dict.unk_token):
                    oov_dict.add_symbol(node_token)
                token_matrix.append([oov_dict.get_symbol_idx(node_token)])
            token_matrix = torch.tensor(
                token_matrix, dtype=torch.long).to(self.device)
            g.node_features['token_id_oov'] = token_matrix
        return oov_dict

    def train_epoch(self, epoch):
        loss_to_print = 0
        for i in range(self.train_data_loader.num_batch):
            self.optimizer.zero_grad()

            input_dict = self.train_data_loader.random_batch()

            batch_graph_list = input_dict['encoder_graph_batch']
            batch_tree_list = input_dict['decoder_tree_batch']
            batch_original_tree_list = input_dict['decoder_output_text_batch']

            oov_dict = self.prepare_ext_vocab(
                batch_graph_list, self.src_vocab) if self.use_copy else None


            if self.use_copy and self.revectorization:
                batch_tree_list_refined = []
                for item in batch_original_tree_list:
                    tgt_list = oov_dict.get_symbol_idx_for_list(item.strip().split())
                    tgt_tree = Tree.convert_to_tree(tgt_list, 0, len(tgt_list), oov_dict)
                    batch_tree_list_refined.append(tgt_tree)
            # for index in range(len(batch_tree_list_refined)):
            #     print("---------------------------------------")
            #     print(batch_tree_list[index])
            #     print(batch_tree_list_refined[index])
            # loss = self.model(batch_graph_list, batch_tree_list, oov_dict=oov_dict)
            loss = self.model(batch_graph_list, batch_tree_list_refined if self.use_copy else batch_tree_list, oov_dict=oov_dict)
            loss.backward()
            torch.nn.utils.clip_grad_value_(
                self.model.parameters(), self.opt["grad_clip"])
            self.optimizer.step()
            loss_to_print += loss
        return loss_to_print/self.train_data_loader.num_batch

    def train(self):
        best_acc = -1

        print("-------------\nStarting training.")
        for epoch in range(1, self.opt["max_epochs"]+1):
            self.model.train()
            loss_to_print = self.train_epoch(epoch)
            # self.scheduler.step()
            print("epochs = {}, train_loss = {:.3f}".format(epoch, loss_to_print))
            # print(self.scheduler.get_lr())
            if epoch > 20 and epoch % 10 == 0:
                # torch.save(checkpoint, "{}/g2t".format(self.checkpoint_dir) + str(i))
                # pickle.dump(checkpoint, open("{}/g2t".format(self.checkpoint_dir) + str(i), "wb"))
                test_acc = self.eval((self.model))
                if test_acc > best_acc:
                    best_acc = test_acc
        print("Best Acc: {:.3f}\n".format(best_acc))

    def eval(self, model):
        device = model.device

        max_dec_seq_length = self.opt["max_dec_seq_length"]
        max_dec_tree_depth = self.opt["max_dec_tree_depth_for_test"]
        
        use_copy = self.test_data_loader.use_copy
        enc_emb_size = model.src_vocab.embedding_dims
        tgt_emb_size = model.tgt_vocab.embedding_dims

        enc_hidden_size = model.decoder.enc_hidden_size
        dec_hidden_size = model.decoder.hidden_size

        model.eval()

        reference_list = []
        candidate_list = []

        data = self.test_data_loader.data

        for i in range(len(data)):
            x = data[i]

            # get input graph list
            input_graph_list = [x[0]]
            # if use_copy:
            oov_dict = self.prepare_ext_vocab(
                input_graph_list, self.test_data_loader.src_vocab)

            # get indexed tgt sequence
            if self.use_copy and self.revectorization:
                reference = oov_dict.get_symbol_idx_for_list(x[1].split())
                # reference = Tree.convert_to_tree(tmp_list, 0, len(tmp_list), oov_dict)
                eval_vocab = oov_dict

            else:
                reference = model.tgt_vocab.get_symbol_idx_for_list(x[1].split())
                eval_vocab = self.test_data_loader.tgt_vocab

            candidate = model.decoder.translate(use_copy,
                                                enc_hidden_size,
                                                dec_hidden_size,
                                                model,
                                                input_graph_list,
                                                self.test_data_loader.src_vocab,
                                                self.test_data_loader.tgt_vocab,
                                                device,
                                                max_dec_seq_length,
                                                max_dec_tree_depth,
                                                oov_dict=oov_dict)
            
            candidate = [int(c) for c in candidate]
            num_left_paren = sum(
                1 for c in candidate if eval_vocab.idx2symbol[int(c)] == "(")
            num_right_paren = sum(
                1 for c in candidate if eval_vocab.idx2symbol[int(c)] == ")")
            diff = num_left_paren - num_right_paren
            if diff > 0:
                for i in range(diff):
                    candidate.append(
                        self.test_data_loader.tgt_vocab.symbol2idx[")"])
            elif diff < 0:
                candidate = candidate[:diff]
            ref_str = convert_to_string(
                reference, eval_vocab)
            cand_str = convert_to_string(
                candidate, eval_vocab)

            # for c in candidate:
            #     if c >= self.test_data_loader.tgt_vocab.vocab_size:
            #         print("====================")
            #         print(oov_dict.symbol2idx)
            #         print(cand_str)
            #         print(ref_str)
            #         print("====================")
            # print(cand_str)
            # print(ref_str)

            reference_list.append(reference)
            candidate_list.append(candidate)
            # print(cand_str)

        test_acc = compute_tree_accuracy(
            candidate_list, reference_list, eval_vocab)
        print("TEST ACCURACY = {:.3f}\n".format(test_acc))
        return test_acc


def convert_to_string(idx_list, form_manager):
    w_list = []
    for i in range(len(idx_list)):
        w_list.append(form_manager.get_idx_symbol(int(idx_list[i])))
    return " ".join(w_list)


def get_split_comma(input_str):
    input_str = input_str.replace(",", " , ")
    input_list = [item.strip() for item in input_str.split()]
    ref_char = "$"
    for index in range(len(input_list)):
        if input_list[index] == ',':
            if input_list[:index].count('(') == input_list[:index].count(')'):
                if input_list[index+1:].count('(') == input_list[index+1:].count(')'):
                    if input_list[index] == ref_char:
                        raise RuntimeError
                    else:
                        input_list[index] = ref_char
    new_str = " ".join(input_list).split('$')
    result_set = set()
    for str_ in new_str:
        result_set.add(str_.strip())
    return result_set


def is_all_same(c1, c2, form_manager):
    all_same = True
    if len(c1) == len(c2):
        all_same = True
        for j in range(len(c1)):
            if c1[j] != c2[j]:
                all_same = False
                break
        if all_same:
            return True
    if len(c1) != len(c2) or all_same == False:
        d1 = " ".join([form_manager.get_idx_symbol(x) for x in c1])
        d2 = " ".join([form_manager.get_idx_symbol(x) for x in c2])
        if get_split_comma(d1) == get_split_comma(d2):
            # print(d1)
            # print(d2)
            # print("============")
            return True
        return False
    raise NotImplementedError("you should not arrive here!")

def compute_accuracy(candidate_list, reference_list, form_manager):
    if len(candidate_list) != len(reference_list):
        print("candidate list has length {}, reference list has length {}\n".format(
            len(candidate_list), len(reference_list)))

    len_min = min(len(candidate_list), len(reference_list))
    c = 0
    for i in range(len_min):
        if is_all_same(candidate_list[i], reference_list[i], form_manager):
            c = c+1
        else:
            pass

    return c/float(len_min)


def compute_tree_accuracy(candidate_list_, reference_list_, form_manager):
    candidate_list = []
    for i in range(len(candidate_list_)):
        candidate_list.append(candidate_list_[i])
    reference_list = []
    for i in range(len(reference_list_)):
        reference_list.append(reference_list_[i])
    return compute_accuracy(candidate_list, reference_list, form_manager)


from .args_jobs import get_args

if __name__ == "__main__":
    start = time.time()
    runner = Jobs(opt=get_args())
    runner.train()
    end = time.time()
    print("total time: {} minutes\n".format((end - start)/60))


import copy
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.nn.init as init

from graph4nlp.pytorch.data.data import from_batch

from .base import Graph2XBase
from graph4nlp.pytorch.modules.prediction.generation.decoder_strategy import DecoderStrategy
from graph4nlp.pytorch.modules.prediction.generation.TreeBasedDecoder import StdTreeDecoder, create_mask
from graph4nlp.pytorch.modules.utils.tree_utils import DataLoaderForGraphEncoder, Tree, Vocab, to_cuda

from graph4nlp.pytorch.data.data import GraphData


class Graph2Tree(Graph2XBase):
    def __init__(self, vocab_model,
                 emb_input_size, 
                 emb_hidden_size, 
                 embedding_style,
                 graph_type, 
                 gnn_direction_option, 
                 gnn_input_size, 
                 gnn_hidden_size, 
                 gnn_output_size,
                 gnn, 
                 gnn_num_layers, 
                 dec_hidden_size,
                 gnn_feat_drop=0.0, 
                 gnn_attn_drop=0.0,
                 emb_fix_word_emb=False, 
                 emb_fix_bert_emb=False, 
                 emb_word_dropout=0.0, 
                 emb_rnn_dropout=0.0,
                 dec_max_decoder_step=50,
                 dec_max_tree_depth=50, 
                 dec_use_copy=False, 
                 dec_use_sibling=False,
                 dec_graph_pooling_strategy="max", 
                 dec_rnn_type="lstm", 
                 dec_attention_type="uniform", 
                 dec_dropout=0.0, 
                 dec_teacher_force_ratio=1.0, 
                 device=None, 
                 **kwargs):
        super(Graph2Tree, self).__init__(vocab_model=vocab_model, 
                                         emb_input_size=emb_input_size,
                                         emb_hidden_size=emb_hidden_size,
                                         graph_type=graph_type, 
                                         gnn_direction_option=gnn_direction_option,
                                         gnn=gnn,
                                         gnn_num_layers=gnn_num_layers,
                                         embedding_style=embedding_style,
                                         device=device,
                                         gnn_feats_dropout=gnn_feat_drop,
                                         gnn_attn_dropout=gnn_attn_drop,
                                         emb_rnn_dropout=emb_rnn_dropout,
                                         emb_fix_word_emb=emb_fix_word_emb,
                                         emb_fix_bert_emb=emb_fix_bert_emb,
                                         emb_word_dropout=emb_word_dropout,
                                         gnn_hidden_size=gnn_hidden_size,
                                         gnn_input_size=gnn_input_size,
                                         gnn_output_size=gnn_output_size,
                                         **kwargs)

        self.use_copy = dec_use_copy
        self.device = device

        self._build_decoder(decoder_length=dec_max_decoder_step,
                            decoder_tree_depth=dec_max_tree_depth,
                            vocab_model=vocab_model,
                            word_emb=self.word_emb,
                            rnn_input_size=emb_hidden_size,
                            input_size=2 * gnn_hidden_size if gnn_direction_option == 'bi_sep' else gnn_hidden_size,
                            hidden_size=dec_hidden_size,
                            use_copy=dec_use_copy,
                            rnn_type=dec_rnn_type,
                            attention_type=dec_attention_type,
                            dec_dropout=dec_dropout,
                            device=device,
                            teacher_force_ratio=dec_teacher_force_ratio,
                            use_sibling=dec_use_sibling,
                            graph_pooling_strategy=dec_graph_pooling_strategy)

    def _build_decoder(self, decoder_length,
                       decoder_tree_depth,
                       vocab_model,
                       word_emb,
                       rnn_input_size,
                       input_size,
                       hidden_size,
                       use_copy=False,
                       rnn_type="lstm",
                       attention_type="uniform",
                       dec_dropout=0.1,
                       device=None,
                       teacher_force_ratio=1.0,
                       use_sibling=False,
                       graph_pooling_strategy="max"):
        
        self.decoder = StdTreeDecoder(attn_type=attention_type, 
                                      embeddings=word_emb,
                                      enc_hidden_size=hidden_size,
                                      dec_emb_size=vocab_model.out_word_vocab.embedding_dims,
                                      dec_hidden_size=hidden_size,
                                      output_size=vocab_model.out_word_vocab.vocab_size,
                                      device=device,
                                      criterion=nn.NLLLoss(size_average=False),
                                      teacher_force_ratio=teacher_force_ratio,
                                      use_sibling=False,
                                      use_attention=True, 
                                      use_copy=use_copy,
                                      use_coverage=False,
                                      fuse_strategy="average",
                                      num_layers=1,
                                      dropout_for_decoder=dec_dropout,
                                      rnn_type=rnn_type,
                                      max_dec_seq_length=decoder_length,
                                      max_dec_tree_depth=decoder_tree_depth,
                                      tgt_vocab=vocab_model.out_word_vocab,
                                      graph_pooling_strategy=graph_pooling_strategy)

    def forward(self, graph_list, tgt_tree_batch, oov_dict=None):
        batch_graph = self.graph_topology(graph_list)
        batch_graph = self.gnn_encoder(batch_graph)
        batch_graph.node_features["rnn_emb"] = batch_graph.node_features['node_feat']

        batch_graph_list_decoder_input = from_batch(batch_graph)
        if self.use_copy and "token_id_oov" not in batch_graph.node_features.keys():
            for g, g_ in zip(batch_graph_list_decoder_input, graph_list):
                g.node_features['token_id_oov'] = g_.node_features['token_id_oov']

        loss = self.decoder(g=batch_graph_list_decoder_input,
                            tgt_tree_batch=tgt_tree_batch, oov_dict=oov_dict)
        return loss

    def init(self, init_weight):
        to_cuda(self.gnn_encoder, self.device)
        to_cuda(self.decoder, self.device)

        print('--------------------------------------------------------------')
        for name, param in self.named_parameters():
            # print(name, param.size())
            if param.requires_grad:
                if ("word_embedding" in name) or ("word_emb_layer" in name) or ("bert_embedding" in name):
                    pass
                else:
                    if len(param.size()) >= 2:
                        if "rnn" in name:
                            init.orthogonal_(param)
                        else:
                            init.xavier_uniform_(param, gain=1.0)
                    else:
                        init.uniform_(param, -init_weight, init_weight)

    @classmethod
    def from_args(cls, opt, vocab_model, device):
        emb_args = cls._get_node_embedding_params(opt)
        gnn_args = cls._get_gnn_params(opt)
        dec_args = cls._get_decoder_params(opt)

        args = (copy.deepcopy(emb_args))
        args.update(gnn_args)
        args.update(dec_args)

        return cls(vocab_model=vocab_model, device=device, **args)

    @staticmethod
    def _get_decoder_params(opt):
        dec_args = opt["decoder_args"]
        shared_args = copy.deepcopy(dec_args["rnn_decoder_share"])
        private_args = copy.deepcopy(dec_args["rnn_decoder_private"])
        ret = copy.deepcopy(dict(shared_args, **private_args))
        dec_ret = {"dec_" + key: value for key, value in ret.items()}
        return dec_ret

    @staticmethod
    def _get_gnn_params(opt):
        args = opt["graph_embedding_args"]
        shared_args = copy.deepcopy(args["graph_embedding_share"])
        private_args = copy.deepcopy(args["graph_embedding_private"])
        if "activation" in private_args.keys():
            private_args["activation"] = getattr(F, private_args["activation"]) if private_args["activation"] else None
        if "norm" in private_args.keys():
            private_args["norm"] = getattr(F, private_args["norm"]) if private_args["norm"] else None

        gnn_shared_args = {"gnn_" + key: value for key, value in shared_args.items()}
        pri_shared_args = {"gnn_" + key: value for key, value in private_args.items()}
        ret = copy.deepcopy(dict(gnn_shared_args, **pri_shared_args))
        ret["gnn"] = opt["graph_embedding_name"]
        return ret

    @staticmethod
    def _get_node_embedding_params(opt):
        args = opt["graph_construction_args"]["node_embedding"]
        ret: dict = copy.deepcopy(args)
        ret.pop("embedding_style")
        emb_ret = {"emb_" + key: value for key, value in ret.items()}
        emb_ret["embedding_style"] = args["embedding_style"]
        emb_ret["graph_type"] = opt["graph_construction_name"]
        return emb_ret



import torch
import torch.nn as nn


class DecoderBase(nn.Module):
    r"""
        The base class for decoder

    Parameters
    ----------
    use_attention: bool, default=True
        Whether to use attention mechanism
    use_copy: bool, default=False
        Whether to use copy attention(pointer network) mechanism
    use_coverage: bool, default=False
        Whether to use coverage attention mechanism
    attention_type: str, option=["uniform", "separate"], default="uniform"
        The attention strategy choice.
        "``uniform``": uniform attention. We will attend on the nodes uniformly.
        "``sep_diff_encoder_type``": separate attention.
                                    We will attend on different encoder results separately.
        "``sep_diff_node_type``": separate attention.
                                    We will attend on different node type separately.
    attention_function_type: str, option=["mlp", "dot", "general"]


    fuse_strategy: str, option=[None, "average", "concatenate"], default=None
        The strategy to fuse attention results generated by separate attention.
        "None": If we do ``uniform`` attention, we will set it to None.
        "``average``": We will take an average on all results.
        "``concatenate``": We will concatenate all results to one.
    """

    def __init__(self, use_attention=True, use_copy=False, use_coverage=False,
                 attention_type="uniform", fuse_strategy="average"):
        super(DecoderBase, self).__init__()
        self.use_attention = use_attention
        self.use_copy = use_copy
        self.use_coverage = use_coverage
        self.attention_type = attention_type
        self.fuse_strategy = fuse_strategy

    def forward(self, **kwargs):
        r"""
            Forward calculation method
        """
        raise NotImplementedError()

    def decode_step(self, **kwargs):
        r"""
            One step for decoding
        """
        raise NotImplementedError()

    def get_decoder_init_state(self, **kwargs):
        r"""
            The initial state for decoding.
        -------

        """
        raise NotImplementedError()


class RNNDecoderBase(DecoderBase):
    def __init__(self, use_attention=True, use_copy=False, use_coverage=False, attention_type="uniform",
                 fuse_strategy="average"):
        super(RNNDecoderBase, self).__init__(use_attention=use_attention, use_copy=use_copy, use_coverage=use_coverage,
                                             attention_type=attention_type, fuse_strategy=fuse_strategy)

    def _build_rnn(self, **kwargs):
        """
            The private function to design the specific RNN models.
            MUST be override by subclass
        """
        raise NotImplementedError()

    def extract_params(self, g):
        raise NotImplementedError()

    def forward(self, g, tgt_seq=None, src_seq=None):
        params = self._extract_params(g)
        return self._run_forward_pass(**params, tgt_seq=tgt_seq)

    def _run_forward_pass(self, graph_node_embedding, graph_node_mask, rnn_node_embedding, graph_level_embedding,
                          graph_edge_embedding=None, graph_edge_mask=None, tgt_seq=None):
        r"""
            The private calculation method for decoder.

        Parameters
        ----------
        tgt_seq: torch.Tensor,
            The target sequence of shape :math:`(B, seq_len)`, where :math:`B`
            is size of batch, :math:`seq_len` is the length of sequences.
            Note that it is consisted by tokens' index.
        graph_node_embedding: torch.Tensor,
            The graph node embedding matrix of shape :math:`(B, N, D_{in})`
        graph_node_mask: torch.Tensor,
            The graph node type mask matrix of shape :math`(B, N)`
        rnn_node_embedding: torch.Tensor,
            The rnn encoded embedding matrix of shape :math`(B, N, D_{in})`
        graph_level_embedding: torch.Tensor,
            graph level embedding of shape :math`(B, D_{in})`
        graph_edge_embedding: torch.Tensor,
            graph edge embedding of shape :math`(B, N, D_{in})`
        graph_edge_mask: torch.Tensor,
            graph edge type embedding

        Returns
        ----------
        logits: torch.Tensor
        attns: torch.Tensor
        """
        raise NotImplementedError()

    def decode_step(self, decoder_input, rnn_state, encoder_out, dec_input_mask, rnn_emb=None,
                    enc_attn_weights_average=None,
                    src_seq=None, oov_dict=None):
        r"""
            One step for decoding
        Parameters
        ----------
        decoder_input: torch.Tensor
            The input for current decoding step
        rnn_state: torch.Tensor
            Rnn_state
        encoder_out: torch.Tensor
            The graph node embedding for decoding
        dec_input_mask: torch.Tensor
            The mask of graph node.
            Notes: ``-1`` is the dummy node, each int larger than -1 is one class for separate attention.
        rnn_emb: torch.Tensor
            The graph node embedding from RNN encoder.
        enc_attn_weights_average: list
            The list of encoder attention weights. It will be used for coverage.
        src_seq: torch.Tensor
            The source sequence. It will be used for copy.
        oov_dict: Vocab
            The vocabulary containing out-of-vocabulary words.
        """
        raise NotImplementedError()

    def get_decoder_init_state(self, rnn_type, batch_size, content=None):
        r"""
            The initial state for RNN decoder.
        Parameters
        ----------
        rnn_type: str, option=["LSTM", "GRU']
            The rnn type.
        batch_size: int
            The batch size of the initial state.
        content: torch.Tensor, default=None
            The initialization of initial state.

        Returns
        -------
            initial_state: Any
        """
        raise NotImplementedError()


class RNNTreeDecoderBase(DecoderBase):
    r"""
        The base class for rnn tree decoder

    Parameters
    ----------
    attentional: bool, default=True
        Whether to use attention mechanism
    use_copy: bool, default=False
        Whether to use copy attention(pointer network) mechanism
    use_coverage: bool, default=False
        Whether to use coverage attention mechanism
    attention_type: str, option=["uniform", "separate"], default="uniform"
        The attention strategy choice.
        "``uniform``": uniform attention. We will attend on the nodes uniformly.
        "``separate_different_encoder_type``": separate attention.
                                               We will attend on different encoder results separately.
        "``separate_different_node_type``": separate attention.
                                            We will attend on different node type separately.

    fuse_strategy: str, option=[None, "average", "concatenate"], default=None
        The strategy to fuse attention results generated by separate attention.
        "None": If we do ``uniform`` attention, we will set it to None.
        "``average``": We will take an average on all results.
        "``concatenate``": We will concatenate all results to one.
    teacher_forcing_ratio: float, default=1.
        The probability rate to use teacher forcing strategy.
    """

    def __init__(self, use_attention=True, use_copy=False, use_coverage=False, attention_type="uniform",
                 fuse_strategy="average"):
        super(RNNTreeDecoderBase, self).__init__(use_attention=use_attention)

    def _build_rnn(self, rnn_type, **kwargs):
        """
            The private function to design the specific RNN models.
            MUST be override by subclass
        """
        raise NotImplementedError()

    def forward(self, encoder_output_dict, tgt_tree_batch, enc_batch):
        return self._run_forward_pass(**encoder_output_dict, tgt_tree_batch=tgt_tree_batch, enc_batch=enc_batch)

    def _run_forward_pass(self, graph_node_embedding, graph_node_mask, rnn_node_embedding, graph_level_embedding,
                          graph_edge_embedding=None, graph_edge_mask=None, tgt=None, enc_batch=None):
        r"""
            The private calculation method for decoder.

        Parameters
        ----------
        tgt: torch.Tensor,
            TODO: add tree decode batch data description
        graph_node_embedding: torch.Tensor,
            The graph node embedding matrix of shape :math:`(B, N, D_{in})`
        graph_node_mask: torch.Tensor,
            The graph node type mask matrix of shape :math`(B, N)`
        rnn_node_embedding: torch.Tensor,
            The rnn encoded embedding matrix of shape :math`(B, N, D_{in})`
        graph_level_embedding: torch.Tensor,
            graph level embedding of shape :math`(B, D_{in})`
        graph_edge_embedding: torch.Tensor,
            graph edge embedding of shape :math`(B, N, D_{in})`
        graph_edge_mask: torch.Tensor,
            graph edge type embedding

        Returns
        ----------
        logits: torch.Tensor
        attns: torch.Tensor
        """
        raise NotImplementedError()



import os

# os.environ["CUDA_VISIBLE_DEVICES"] = "2"

# os.environ['CUDA_LAUNCH_BLOCKING'] = "5"
from graph4nlp.pytorch.data.data import GraphData
from graph4nlp.pytorch.datasets.kinship import KinshipDataset, KinshipTestDataset
# from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
# from graph4nlp.pytorch.modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
# from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
from graph4nlp.pytorch.modules.graph_embedding.gat import GAT
from graph4nlp.pytorch.modules.graph_embedding.ggnn import GGNN
from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
from graph4nlp.pytorch.modules.prediction.classification.kg_completion.DistMult import DistMult, DistMultLayer

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.optim as optim
import dgl
import numpy as np

from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
from graph4nlp.pytorch.modules.graph_construction.embedding_construction import EmbeddingConstruction
from graph4nlp.pytorch.modules.loss.kg_loss import *

class RankingAndHits(EvaluationMetricBase):
    def __init__(self, model_path='best_graph2distmult', batch_size=128):
        super(RankingAndHits, self).__init__()
        self.batch_size = batch_size
        self.best_mrr = 0.
        self.best_hits1 = 0.
        self.best_hits10 = 0.
        self.model_path = model_path

    def calculate_scores(self, model, dataloader, name, kg_graph, device):
        print('')
        print('-' * 50)
        print(name)
        hits_left = []
        hits_right = []
        hits = []
        ranks = []
        ranks_left = []
        ranks_right = []
        for i in range(10):
            hits_left.append([])
            hits_right.append([])
            hits.append([])
        with open('output_model.txt', 'w') as file:
            for data in dataloader:
                e1, rel, e2_multi, e2_multi_tensor_idx, \
                e2, rel_eval, e1_multi, e1_multi_tensor_idx = data

                e1 = e1.view(-1, 1).to(device)
                rel = rel.view(-1, 1).to(device)
                e2_multi = e2_multi.to(device)
                e2_multi_tensor_idx = e2_multi_tensor_idx.to(device)

                e2 = e2.view(-1, 1).to(device)
                rel_eval = rel_eval.view(-1, 1).to(device)
                e1_multi = e1_multi.to(device)
                e1_multi_tensor_idx = e1_multi_tensor_idx.to(device)
                # if model.loss_name == "SoftplusLoss" or model.loss_name == "SigmoidLoss":
                #     pred1, pos1, neg1 = model(kg_graph, e1, rel, e2_multi, require_loss=False)
                #     pred2, pos2, neg2 = model(kg_graph, e2, rel_eval, e1_multi, require_loss=False)
                # else:
                pred1 = model(kg_graph, e1, rel, e2_multi, require_loss=False)
                pred2 = model(kg_graph, e2, rel_eval, e1_multi, require_loss=False)
                pred1, pred2 = pred1.data, pred2.data
                e1, e2 = e1.data, e2.data
                e2_multi1, e2_multi2 = e2_multi_tensor_idx.data, e1_multi_tensor_idx.data

                if e1.size()[0]!=self.batch_size:
                    continue

                # if model.loss_name == "SoftplusLoss" or model.loss_name == "SigmoidLoss":
                #     e2_multi1_binary = str2var['e2_multi1_binary'].float()
                #     e2_multi2_binary = str2var['e2_multi2_binary'].float()
                #     pred1, pos1, neg1 = model.forward(e1, rel, X, adjacencies, e2_multi1_binary)
                #     pred2, pos2, neg2 = model.forward(e2, rel_reverse, X, adjacencies, e2_multi2_binary)
                # else:
                #     pred1 = model.forward(e1, rel, X, adjacencies)
                #     pred2 = model.forward(e2, rel_reverse, X, adjacencies)
                # pred1, pred2 = pred1.data, pred2.data
                # e1, e2 = e1.data, e2.data
                # e2_multi1, e2_multi2 = e2_multi1.data, e2_multi2.data
                for i in range(self.batch_size):
                    # these filters contain ALL labels
                    filter1 = e2_multi1[i].long()
                    filter2 = e2_multi2[i].long()

                    num = e1[i, 0].item()
                    # save the prediction that is relevant
                    target_value1 = pred1[i,e2.cpu().numpy()[i, 0].item()].item()
                    target_value2 = pred2[i,e1.cpu().numpy()[i, 0].item()].item()
                    # zero all known cases (this are not interesting)
                    # this corresponds to the filtered setting
                    pred1[i][filter1] = 0.0
                    pred2[i][filter2] = 0.0
                    # write base the saved values
                    pred1[i][e2[i]] = target_value1
                    pred2[i][e1[i]] = target_value2

                    # print(e1[i, 0])


                # sort and rank
                max_values1, argsort1 = torch.sort(pred1, 1, descending=True)
                max_values2, argsort2 = torch.sort(pred2, 1, descending=True)

                argsort1 = argsort1.cpu().numpy()
                argsort2 = argsort2.cpu().numpy()
                for i in range(self.batch_size):
                    # find the rank of the target entities
                    rank1 = np.where(argsort1[i]==e2.cpu().numpy()[i, 0])[0][0]
                    if model.loss_name in ['SoftplusLoss', 'SigmoidLoss'] and max_values1[i][rank1] == max_values1[i][0]:
                        rank1 = 0
                    rank2 = np.where(argsort2[i]==e1.cpu().numpy()[i, 0])[0][0]
                    if model.loss_name in ['SoftplusLoss', 'SigmoidLoss'] and max_values2[i][rank2] == max_values2[i][0]:
                        rank2 = 0
                    # rank+1, since the lowest rank is rank 1 not rank 0
                    ranks.append(rank1 + 1)
                    ranks_left.append(rank1 + 1)
                    ranks.append(rank2 + 1)
                    ranks_right.append(rank2 + 1)

                    file.write(str(e2.cpu().numpy()[i, 0].item()) + '\t')
                    file.write(str(rank1 + 1) + '\n')
                    file.write(str(e1.cpu().numpy()[i, 0].item()) + '\t')
                    file.write(str(rank2 + 1) + '\n')

                    # this could be done more elegantly, but here you go
                    for hits_level in range(10):
                        if rank1 <= hits_level:
                            hits[hits_level].append(1.0)
                            hits_left[hits_level].append(1.0)
                        else:
                            hits[hits_level].append(0.0)
                            hits_left[hits_level].append(0.0)

                        if rank2 <= hits_level:
                            hits[hits_level].append(1.0)
                            hits_right[hits_level].append(1.0)
                        else:
                            hits[hits_level].append(0.0)
                            hits_right[hits_level].append(0.0)

        for i in range(10):
            print('Hits left @{0}: {1}'.format(i + 1, np.mean(hits_left[i])))
            print('Hits right @{0}: {1}'.format(i + 1, np.mean(hits_right[i])))
            print('Hits @{0}: {1}'.format(i + 1, np.mean(hits[i])))
        print('Mean rank left: ', np.mean(ranks_left))
        print('Mean rank right: ', np.mean(ranks_right))
        print('Mean rank: ', np.mean(ranks))
        print('Mean reciprocal rank left: ', np.mean(1. / np.array(ranks_left)))
        print('Mean reciprocal rank right: ', np.mean(1. / np.array(ranks_right)))
        print('Mean reciprocal rank: ', np.mean(1. / np.array(ranks)))

        if np.mean(hits[0])>self.best_hits1:
            self.best_hits1 = np.mean(hits[0])

        if np.mean(hits[9])>self.best_hits10:
            self.best_hits10 = np.mean(hits[9])

        if np.mean(1. / np.array(ranks))>self.best_mrr:
            self.best_mrr = np.mean(1. / np.array(ranks))
            print('saving best model...')
            torch.save(model.state_dict(), self.model_path)

        print('-' * 50)
        print('')

# TODO: 1. initialize graph.node_features['edge_feat']/self.distmult.rel_emb with self.embedding_layer
# TODO: 2. learn graph.node_features['edge_emb'] from GNN (edge2node)

class Graph2DistMult(nn.Module):
    def __init__(self, vocab, num_entities, hidden_size=300, num_relations=None, direction_option='uni', loss_name='BCELoss'):
        super(Graph2DistMult, self).__init__()
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        self.vocab = vocab
        self.num_entities = num_entities

        self.node_emb = nn.Embedding(num_entities, hidden_size)

        embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
                           'seq_info_encode_strategy': "none"}

        self.embedding_layer = EmbeddingConstruction(self.vocab.in_word_vocab,
                                                     embedding_style['word_emb_type'],
                                                     embedding_style['node_edge_emb_strategy'],
                                                     embedding_style['seq_info_encode_strategy'],
                                                     hidden_size=hidden_size,
                                                     fix_word_emb=False,
                                                     dropout=0.2,
                                                     device=self.device)

        # self.graph_topology = IEBasedGraphConstruction(vocab=self.vocab.in_word_vocab,
        #                                                embedding_style=embedding_style,
        #                                                use_cuda=False)
        # output_graph0 = iegc.embedding(output_graph0)


        # self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
        #                                                        vocab=vocab.in_word_vocab,
        #                                                        hidden_size=hidden_size, dropout=0.2, use_cuda=True,
        #                                                        fix_word_emb=False)

        # self.word_emb = self.graph_topology.embedding_layer.word_emb_layers[0].word_emb_layer

        # self.gnn_encoder = GAT(2, hidden_size, hidden_size, hidden_size, [2, 1], direction_option=direction_option)
        self.gnn_encoder = GGNN(1, hidden_size, hidden_size, direction_option=direction_option)
        self.bn = torch.nn.BatchNorm1d(hidden_size)  # necessary for this task

        self.distmult = DistMult(rel_emb_from_gnn=False, num_relations=num_relations,
                                 embedding_dim=hidden_size, loss_name=loss_name)

        self.loss_name = loss_name
        if loss_name == 'BCELoss':
            self.loss = torch.nn.BCELoss()
        elif loss_name == "SoftplusLoss":
            self.loss = SoftplusLoss()
        elif loss_name == "SigmoidLoss":
            self.loss = SigmoidLoss()
        elif loss_name == "SoftMarginLoss":
            self.loss = nn.SoftMarginLoss()
        elif loss_name == "MSELoss":
            self.loss = nn.MSELoss()
        else:
            raise NotImplementedError()

        self.reset_parameters()

    # def embedding(self, graph: GraphData):
    #     node_attributes = graph.node_attributes
    #     edge_attributes = graph.edge_attributes
    #
    #     # Build embedding(initial feature vector) for graph nodes.
    #     # Each node may contains multiple tokens.
    #     node_idxs_list = []
    #     node_len_list = []
    #     for node_id, node_dict in node_attributes.items():
    #         node_word_idxs = []
    #         for token in node_dict['token'].split():
    #             node_word_idxs.append(self.vocab.in_word_vocab.getIndex(token))
    #         node_idxs_list.append(node_word_idxs)
    #         node_len_list.append(len(node_word_idxs))
    #     max_size = max(node_len_list)
    #     node_idxs_list = [x+[self.vocab.in_word_vocab.PAD]*(max_size-len(x)) for x in node_idxs_list]
    #     node_idxs_tensor = torch.LongTensor(node_idxs_list).to(self.device)
    #     # if self.embedding_layer.node_edge_emb_strategy == 'mean':
    #     #     node_len_tensor = torch.LongTensor(node_len_list).view(-1, 1)
    #     # else:
    #     node_len_tensor = torch.LongTensor(node_len_list).to(self.device)
    #     num_nodes = torch.LongTensor([len(node_len_list)]).to(self.device)
    #     node_feat = self.embedding_layer(node_idxs_tensor, node_len_tensor, num_nodes)
    #     graph.node_features['node_feat'] = node_feat
    #
    #     if 'token' in edge_attributes[0].keys():
    #         # If edge information is stored in `edge_attributes`,
    #         # build embedding(initial feature vector) for graph edges.
    #         # Each edge may contains multiple tokens.
    #         edge_idxs_list = []
    #         edge_len_list = []
    #         for edge_id, edge_dict in edge_attributes.items():
    #             edge_word_idxs = []
    #             for token in edge_dict['token'].split():
    #                 edge_word_idxs.append(self.vocab.in_word_vocab.getIndex(token))
    #             edge_idxs_list.append(edge_word_idxs)
    #             edge_len_list.append(len(edge_word_idxs))
    #
    #         max_size = max(edge_len_list)
    #         edge_idxs_list = [x + [self.vocab.in_word_vocab.PAD] * (max_size - len(x)) for x in edge_idxs_list]
    #         edge_idxs_tensor = torch.LongTensor(edge_idxs_list).to(self.device)
    #         # if self.embedding_layer.node_edge_emb_strategy == 'mean':
    #         #     edge_len_tensor = torch.LongTensor(edge_len_list).view(-1, 1)
    #         # else:
    #         edge_len_tensor = torch.LongTensor(edge_len_list).to(self.device)
    #         num_edges = torch.LongTensor([len(edge_len_list)]).to(self.device)
    #         edge_feat = self.embedding_layer(edge_idxs_tensor, edge_len_tensor, num_edges)
    #         graph.edge_features['edge_feat'] = edge_feat
    #
    #     return graph

    def embedding(self, kg_graph):
        # graph_nodes_idx = [self.vocab.in_word_vocab.word2index[token]
        #                    for x in kg_graph.graph_attributes['graph_nodes']
        #                    for token in x.split(' ')]
        graph_nodes_idx = []
        node_len_list = []
        for x in kg_graph.graph_attributes['graph_nodes']:
            tmp_list = []
            for token in x.split(' '):
                tmp_list.append(self.vocab.in_word_vocab.word2index[token])

            graph_nodes_idx.append(tmp_list)
            node_len_list.append(len(tmp_list))

        max_size = max(node_len_list)
        graph_nodes_idx = [x + [self.vocab.in_word_vocab.PAD] * (max_size - len(x)) for x in graph_nodes_idx]
        graph_nodes_idx = torch.tensor(graph_nodes_idx, dtype=torch.long).to(self.device)
        node_len_tensor = torch.tensor(node_len_list, dtype=torch.long).to(self.device)
        num_nodes = torch.tensor([len(node_len_list)], dtype=torch.long).to(self.device)
        node_feat = self.embedding_layer(graph_nodes_idx, node_len_tensor, num_nodes)
        return node_feat

    def reset_parameters(self):
        nn.init.xavier_normal_(self.node_emb.weight.data)
        # xavier_normal_(self.emb_rel.weight.data)

    def forward(self, kg_graph, e1, rel, e2_multi=None, require_loss=True):
        # kg_graph.node_features['node_feat'] = self.node_emb.weight

        # kg_graph.node_features['node_emb'] = self.node_emb.weight
        kg_graph.node_features['node_feat'] = self.embedding(kg_graph)

        list_e_r_pair_idx = list(zip(e1.squeeze().tolist(), rel.squeeze().tolist()))

        # run GNN
        kg_graph = self.gnn_encoder(kg_graph)
        kg_graph.node_features['node_emb'] = self.bn(kg_graph.node_features['node_emb'])
        kg_graph.graph_attributes['list_e_r_pair_idx'] = list_e_r_pair_idx
        kg_graph.graph_attributes['multi_binary_label'] = e2_multi

        # down-task
        kg_graph = self.distmult(kg_graph)
        if require_loss:
            if self.loss_name == "SoftplusLoss" or self.loss_name == "SigmoidLoss":
                loss = self.loss(kg_graph.graph_attributes['p_score'],
                                 kg_graph.graph_attributes['n_score'])
            else:
                loss = self.loss(kg_graph.graph_attributes['logits'], e2_multi)
            return kg_graph.graph_attributes['logits'], loss
        else:
            return kg_graph.graph_attributes['logits']


class Kinship:
    def __init__(self):
        super(Kinship, self).__init__()
        self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
        self._build_dataloader()
        self._build_model()
        self._build_optimizer()
        self._build_evaluation()

    def _build_dataloader(self):
        # train_set = KinshipDataset(root_dir="graph4nlp/pytorch/test/dataset/kinship",
        # train_set = KinshipDataset(root_dir="/mnt/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
        train_set = KinshipDataset(root_dir="/Users/gaohanning/PycharmProjects/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
                                  topology_builder=None,
                                  topology_subdir='e1rel_to_e2',
                                  share_vocab=True)

        # TODO: When edge2node is True, num_entities != train_set.KG_graph.get_node_num()
        self.num_entities = train_set.KG_graph.graph_attributes['num_entities']
        self.num_relations = train_set.KG_graph.graph_attributes['num_relations']

        self.kg_graph = train_set.KG_graph

        test_set = KinshipTestDataset(
            root_dir="/Users/gaohanning/PycharmProjects/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
            # root_dir="/mnt/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
            topology_builder=None,
            topology_subdir='e1rel_to_e2',
            share_vocab=True,
            KG_graph=self.kg_graph)

        self.train_dataloader = DataLoader(train_set, batch_size=128, shuffle=True,
                                           num_workers=1,
                                           collate_fn=train_set.collate_fn)
        self.test_dataloader = DataLoader(test_set, batch_size=128, shuffle=True,
                                          num_workers=1,
                                          collate_fn=test_set.collate_fn)
        self.vocab = train_set.vocab_model

    def _build_model(self):
        self.model = Graph2DistMult(self.vocab,
                                    num_entities=self.num_entities,
                                    num_relations=self.num_relations,
                                    loss_name='SigmoidLoss').to(self.device)

    def _build_optimizer(self):
        parameters = [p for p in self.model.parameters() if p.requires_grad]
        self.optimizer = optim.Adam(parameters, lr=1e-3)

    def _build_evaluation(self):
        self.metrics = [RankingAndHits()]

    def train(self):
        for epoch in range(200):
            self.model.train()
            loss_list = []

            for data in self.train_dataloader:
                e1, rel, e2_multi, e2_multi_tensor_idx = data
                e1 = e1.to(self.device)
                rel = rel.to(self.device)
                e2_multi = e2_multi.to(self.device)
                _, loss = self.model(self.kg_graph, e1, rel, e2_multi, require_loss=True)
                loss_list.append(loss.item())
                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()

            print('#' * 50)
            print("Epoch: {}".format(epoch))
            print('train loss =' + str(sum(loss_list)/len(loss_list)))
            print('#' * 50)
            self.evaluate()

    def evaluate(self):
        self.model.eval()

        self.metrics[0].calculate_scores(self.model,
                                         self.test_dataloader,
                                         name='test',
                                         kg_graph=self.kg_graph,
                                         device=self.device)


if __name__ == "__main__":
    runner = Kinship()
    max_score = runner.train()
    print("Train finish, best score: {:.3f}".format(max_score))

# import os
#
# os.environ["CUDA_VISIBLE_DEVICES"] = "2"
#
# os.environ['CUDA_LAUNCH_BLOCKING'] = "5"
# from graph4nlp.pytorch.data.data import GraphData
# from graph4nlp.pytorch.datasets.kinship import KinshipDataset, KinshipTestDataset
# # from graph4nlp.pytorch.modules.graph_construction.dependency_graph_construction import DependencyBasedGraphConstruction
# # from graph4nlp.pytorch.modules.graph_construction.ie_graph_construction import IEBasedGraphConstruction
# # from graph4nlp.pytorch.modules.prediction.generation.StdRNNDecoder import StdRNNDecoder
# from graph4nlp.pytorch.modules.graph_construction.gat import GAT
# from graph4nlp.pytorch.modules.utils.vocab_utils import Vocab
# from graph4nlp.pytorch.modules.prediction.classification.kg_completion.DistMult import DistMult
#
# import torch
# import torch.nn as nn
# from torch.utils.data import DataLoader
# import torch.optim as optim
# import dgl
#
# from graph4nlp.pytorch.modules.evaluation.base import EvaluationMetricBase
# from graph4nlp.pytorch.modules.graph_construction.embedding_construction import EmbeddingConstruction
# from graph4nlp.pytorch.modules.loss.kg_loss import *
#
# class ExpressionAccuracy(EvaluationMetricBase):
#     def __init__(self):
#         super(ExpressionAccuracy, self).__init__()
#
#     def calculate_scores(self, ground_truth, predict):
#         correct = 0
#         assert len(ground_truth) == len(predict)
#         for gt, pred in zip(ground_truth, predict):
#             print("ground truth: ", gt)
#             print("prediction: ", pred)
#
#             if gt == pred:
#                 correct += 1.
#         return correct / len(ground_truth)
#
#
# def logits2seq(prob: torch.Tensor):
#     ids = prob.argmax(dim=-1)
#     return ids
#
#
# def wordid2str(word_ids, vocab: Vocab):
#     ret = []
#     assert len(word_ids.shape) == 2, print(word_ids.shape)
#     for i in range(word_ids.shape[0]):
#         id_list = word_ids[i, :]
#         ret_inst = []
#         for j in range(id_list.shape[0]):
#             if id_list[j] == vocab.EOS:
#                 break
#             token = vocab.getWord(id_list[j])
#             ret_inst.append(token)
#         ret.append(" ".join(ret_inst))
#     return ret
#
#
# class Graph2seqLoss(nn.Module):
#     def __init__(self, vocab: Vocab):
#         super(Graph2seqLoss, self).__init__()
#         self.loss_func = nn.NLLLoss()
#         self.VERY_SMALL_NUMBER = 1e-31
#         self.vocab = vocab
#
#     def forward(self, prob, gt):
#         assert prob.shape[0:1] == gt.shape[0:1]
#         assert len(prob.shape) == 3
#         log_prob = torch.log(prob + self.VERY_SMALL_NUMBER)
#         batch_size = gt.shape[0]
#         step = gt.shape[1]
#
#         mask = 1 - gt.data.eq(self.vocab.PAD).float()
#
#         prob_select = torch.gather(log_prob.view(batch_size * step, -1), 1, gt.view(-1, 1))
#
#         prob_select_masked = - torch.masked_select(prob_select, mask.view(-1, 1).byte())
#         loss = torch.mean(prob_select_masked)
#         return loss
#
#
# class Graph2DistMult(nn.Module):
#     def __init__(self, vocab, hidden_size=300, num_relations=None, direction_option='bi_fuse', loss_name='BCELoss'):
#         super(Graph2DistMult, self).__init__()
#         self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
#         self.vocab = vocab
#         embedding_style = {'word_emb_type': 'w2v', 'node_edge_emb_strategy': "mean",
#                            'seq_info_encode_strategy': "none"}
#
#         self.embedding_layer = EmbeddingConstruction(self.vocab.in_word_vocab,
#                                                      embedding_style['word_emb_type'],
#                                                      embedding_style['node_edge_emb_strategy'],
#                                                      embedding_style['seq_info_encode_strategy'],
#                                                      hidden_size=hidden_size,
#                                                      fix_word_emb=False,
#                                                      dropout=0.2,
#                                                      device=self.device)
#
#         # self.graph_topology = IEBasedGraphConstruction(vocab=self.vocab.in_word_vocab,
#         #                                                embedding_style=embedding_style,
#         #                                                use_cuda=False)
#         # output_graph0 = iegc.embedding(output_graph0)
#
#
#         # self.graph_topology = DependencyBasedGraphConstruction(embedding_style=embedding_style,
#         #                                                        vocab=vocab.in_word_vocab,
#         #                                                        hidden_size=hidden_size, dropout=0.2, use_cuda=True,
#         #                                                        fix_word_emb=False)
#
#         # self.word_emb = self.graph_topology.embedding_layer.word_emb_layers[0].word_emb_layer
#         self.gnn_encoder = GAT(2, hidden_size, hidden_size, hidden_size, [2, 1], direction_option=direction_option)
#         self.distmult = DistMult(rel_emb_from_gnn=False, num_relations=num_relations, embedding_dim=hidden_size)
#         # self.loss_calc = Graph2seqLoss(self.vocab.in_word_vocab)
#
#         self.loss_name = loss_name
#         if loss_name == 'BCELoss':
#             self.loss = torch.nn.BCELoss()
#         elif loss_name == "SoftplusLoss":
#             self.loss = SoftplusLoss()
#         elif loss_name == "SigmoidLoss":
#             self.loss = SigmoidLoss()
#         elif loss_name == "SoftMarginLoss":
#             self.loss = nn.SoftMarginLoss()
#         elif loss_name == "MSELoss":
#             self.loss = nn.MSELoss()
#         else:
#             raise NotImplementedError()
#
#     def embedding(self, graph: GraphData):
#         node_attributes = graph.node_attributes
#         edge_attributes = graph.edge_attributes
#
#         # Build embedding(initial feature vector) for graph nodes.
#         # Each node may contains multiple tokens.
#         node_idxs_list = []
#         node_len_list = []
#         for node_id, node_dict in node_attributes.items():
#             node_word_idxs = []
#             for token in node_dict['token'].split():
#                 node_word_idxs.append(self.vocab.in_word_vocab.getIndex(token))
#             node_idxs_list.append(node_word_idxs)
#             node_len_list.append(len(node_word_idxs))
#         max_size = max(node_len_list)
#         node_idxs_list = [x+[self.vocab.in_word_vocab.PAD]*(max_size-len(x)) for x in node_idxs_list]
#         node_idxs_tensor = torch.LongTensor(node_idxs_list)
#         # if self.embedding_layer.node_edge_emb_strategy == 'mean':
#         #     node_len_tensor = torch.LongTensor(node_len_list).view(-1, 1)
#         # else:
#         node_len_tensor = torch.LongTensor(node_len_list)
#         num_nodes = torch.LongTensor([len(node_len_list)])
#         node_feat = self.embedding_layer(node_idxs_tensor, node_len_tensor, num_nodes)
#         graph.node_features['node_feat'] = node_feat
#
#         if 'token' in edge_attributes[0].keys():
#             # If edge information is stored in `edge_attributes`,
#             # build embedding(initial feature vector) for graph edges.
#             # Each edge may contains multiple tokens.
#             edge_idxs_list = []
#             edge_len_list = []
#             for edge_id, edge_dict in edge_attributes.items():
#                 edge_word_idxs = []
#                 for token in edge_dict['token']:
#                     edge_word_idxs.append(self.vocab.in_word_vocab.getIndex(token))
#                 edge_idxs_list.append(edge_word_idxs)
#                 edge_len_list.append(len(edge_word_idxs))
#
#             max_size = max(edge_len_list)
#             edge_idxs_list = [x + [self.vocab.in_word_vocab.PAD] * (max_size - len(x)) for x in edge_idxs_list]
#             edge_idxs_tensor = torch.LongTensor(edge_idxs_list)
#             # if self.embedding_layer.node_edge_emb_strategy == 'mean':
#             #     edge_len_tensor = torch.LongTensor(edge_len_list).view(-1, 1)
#             # else:
#             edge_len_tensor = torch.LongTensor(edge_len_list)
#             num_edges = torch.LongTensor([len(edge_len_list)])
#             edge_feat = self.embedding_layer(edge_idxs_tensor, edge_len_tensor, num_edges)
#             graph.edge_features['edge_feat'] = edge_feat
#
#         return graph
#
#     def forward(self, kg_graph, e1, rel, e2_multi=None, require_loss=True):
#         kg_graph = self.embedding(kg_graph)
#         # batch_dgl_graph = self.graph_topology(graph_list)
#         # do graph nn here
#         # convert DGLGraph to GraphData
#         # batch_graph = GraphData()
#         # batch_graph.from_dgl(kg_graph)
#         # kg_graph_dgl = kg_graph.to_dgl()
#
#         list_e_r_pair_idx = list(zip(e1.squeeze().tolist(), rel.squeeze().tolist()))
#
#         # run GNN
#         batch_graph = self.gnn_encoder(kg_graph)
#         batch_graph.graph_attributes['list_e_r_pair_idx'] = list_e_r_pair_idx
#         # kg_graph_dgl.ndata['node_emb'] = kg_graph.node_features['node_emb']
#         # kg_graph_dgl.ndata['rnn_emb'] = kg_graph.node_features['node_feat']
#
#         # dgl_graph_list = dgl.unbatch(kg_graph_dgl)
#         # for g, dg in zip(graph_list, dgl_graph_list):
#         #     g.node_features["node_emb"] = dg.ndata["node_emb"]
#         #     g.node_features["rnn_emb"] = dg.ndata["rnn_emb"]
#
#         # down-task
#         batch_graph = self.distmult(batch_graph)
#         if require_loss:
#             loss = self.loss(batch_graph.graph_attributes['logits'], e2_multi)
#             return batch_graph.graph_attributes['logits'], loss
#         else:
#             return batch_graph.graph_attributes['logits']
#
#
# class Kinship:
#     def __init__(self):
#         super(Kinship, self).__init__()
#         self.device = 'cuda' if torch.cuda.is_available() else 'cpu'
#         self._build_dataloader()
#         self._build_model()
#         self._build_optimizer()
#         self._build_evaluation()
#
#     def _build_dataloader(self):
#         # train_set = KinshipDataset(root_dir="graph4nlp/pytorch/test/dataset/kinship",
#         train_set = KinshipDataset(root_dir="/Users/gaohanning/PycharmProjects/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
#                                   topology_builder=None,
#                                   topology_subdir='e1rel_to_e2',
#                                   share_vocab=True)
#
#         # TODO: When edge2node is True, num_entities != train_set.KG_graph.get_node_num()
#         self.num_entities = train_set.KG_graph.graph_attributes['num_entities']
#         self.num_relations = train_set.KG_graph.graph_attributes['num_relations']
#
#         self.kg_graph = train_set.KG_graph
#
#         test_set = KinshipTestDataset(
#             root_dir="/Users/gaohanning/PycharmProjects/graph4nlp/graph4nlp/pytorch/test/dataset/kinship",
#             topology_builder=None,
#             topology_subdir='e1rel_to_e2',
#             share_vocab=True)
#
#         self.train_dataloader = DataLoader(train_set, batch_size=24, shuffle=True,
#                                            num_workers=1,
#                                            collate_fn=train_set.collate_fn)
#         self.test_dataloader = DataLoader(test_set, batch_size=24, shuffle=True,
#                                           num_workers=1,
#                                           collate_fn=test_set.collate_fn)
#         self.vocab = train_set.vocab_model
#
#     def _build_model(self):
#         self.model = Graph2DistMult(self.vocab, num_relations=self.num_relations).to(self.device)
#
#     def _build_optimizer(self):
#         parameters = [p for p in self.model.parameters() if p.requires_grad]
#         self.optimizer = optim.Adam(parameters, lr=1e-3)
#
#     def _build_evaluation(self):
#         self.metrics = [ExpressionAccuracy()]
#
#     def train(self):
#         max_score = -1
#         for epoch in range(200):
#             self.model.train()
#             print("Epoch: {}".format(epoch))
#             for data in self.train_dataloader:
#                 e1, rel, e2_multi = data
#                 e1 = e1.to(self.device)
#                 rel = rel.to(self.device)
#                 e2_multi = e2_multi.to(self.device)
#                 _, loss = self.model(self.kg_graph, e1, rel, e2_multi, require_loss=True)
#                 print(loss.item())
#                 self.optimizer.zero_grad()
#                 loss.backward()
#                 self.optimizer.step()
#
#                 # self.evaluate()
#             if epoch > 4:
#                 self.evaluate()
#                 # score = self.evaluate()
#                 # max_score = max(max_score, score)
#         return max_score
#
#     def evaluate(self):
#         self.model.eval()
#         pred_collect = []
#         gt_collect = []
#         acc_list = []
#         for data in self.test_dataloader:
#             e1, rel, e2_multi = data
#             e1 = e1.to(self.device)
#             rel = rel.to(self.device)
#             e2_multi = e2_multi.to(self.device)
#
#             prob = self.model(self.kg_graph, e1, rel, e2_multi, require_loss=False)
#             acc = torch.sum(torch.gt(prob, 0.5)==e2_multi).item()/(self.test_dataloader.batch_size*self.kg_graph.graph_attributes['num_entities'])
#             # print('acc='+str(acc))
#             # break
#             acc_list.append(acc)
#         print('acc='+str(sum(acc_list)/len(acc_list)))
#             # pred = logits2seq(prob)
#             #
#             # pred_str = wordid2str(pred.detach().cpu(), self.vocab.in_word_vocab)
#             # tgt_str = wordid2str(tgt, self.vocab.in_word_vocab)
#             # pred_collect.extend(pred_str)
#             # gt_collect.extend(tgt_str)
#
#         # score = self.metrics[0].calculate_scores(ground_truth=gt_collect, predict=pred_collect)
#         # print("accuracy: {:.3f}".format(score))
#         # return score
#
#
# if __name__ == "__main__":
#     runner = Kinship()
#     max_score = runner.train()
#     print("Train finish, best score: {:.3f}".format(max_score))



import argparse, time
import numpy as np
import networkx as nx
import torch
import torch.nn as nn
import torch.nn.functional as F
from dgl import DGLGraph
from dgl.data import register_data_args, load_data
from prediction.classification.node_classification.FeedForwardNNLayer import FeedForwardNNLayer  
from dgl.nn.pytorch import GraphConv
from ...modules.prediction.classification.node_classification.BiLSTMFeedForwardNNLayer import BiLSTMFeedForwardNNLayer 
from ...data.data import *

class GCN(nn.Module):
    def __init__(self,
                 in_feats,
                 n_hidden,
                 n_classes,
                 n_layers,
                 aggreagte_type,
                 output_size,
                 dropout):
        super(GCN, self).__init__()
        self.layers = nn.ModuleList()
        self.model = GraphSAGE(n_layers,
                    in_feats,
                    n_hidden,
                    output_size,
                    aggreagte_type,
                    direction_option=args.direction_option,
                    feat_drop=0.6,
                    bias=True,
                    activation=nn.ReLU)      
        # output layer            
        #self.classifier=FeedForwardNN(output_size,n_classes,[16],nn.Sigmoid())
        self.classifier=BiLSTMFeedForwardNN(output_size,n_classes,16)   
        self.dropout = nn.Dropout(p=dropout)



    def forward(self, graph):
        out_graph = self.model(graph) 
        out_graph_=self.classifier(out_graph)       
        return out_graph_.node_features['logits']


def evaluate(model, G, labels, mask):
    model.eval()
    with torch.no_grad():
        logits = model(G)
        logits = logits[mask]
        labels = labels[mask]
        _, indices = torch.max(logits, dim=1)
        correct = torch.sum(indices == labels)
        return correct.item() * 1.0 / len(labels)



def main(args):
    # load and preprocess dataset
    data = load_data(args)
    features = torch.FloatTensor(data.features)
    labels = torch.LongTensor(data.labels)
    if hasattr(torch, 'BoolTensor'):
        train_mask = torch.BoolTensor(data.train_mask)
        val_mask = torch.BoolTensor(data.val_mask)
        test_mask = torch.BoolTensor(data.test_mask)
    else:
        train_mask = torch.ByteTensor(data.train_mask)
        val_mask = torch.ByteTensor(data.val_mask)
        test_mask = torch.ByteTensor(data.test_mask)
    in_feats = features.shape[1]
    n_classes = data.num_labels
    n_edges = data.graph.number_of_edges()
    print("""----Data statistics------'
      #Edges %d
      #Classes %d
      #Train samples %d
      #Val samples %d
      #Test samples %d""" %
          (n_edges, n_classes,
              train_mask.int().sum().item(),
              val_mask.int().sum().item(),
              test_mask.int().sum().item()))


    if args.gpu < 0:
        cuda = False
    else:
        cuda = True
        torch.cuda.set_device(args.gpu)
        features = features.cuda()
        labels = labels.cuda()
        train_mask = train_mask.cuda()
        val_mask = val_mask.cuda()
        test_mask = test_mask.cuda()


    # graph preprocess and calculate normalization factor
    g = data.graph
    # add self loop
    if args.self_loop:
        g.remove_edges_from(nx.selfloop_edges(g))
        g.add_edges_from(zip(g.nodes(), g.nodes()))
    g = DGLGraph(g)
    n_edges = g.number_of_edges()
    # normalization
    degs = g.in_degrees().float()
    norm = torch.pow(degs, -0.5)
    norm[torch.isinf(norm)] = 0
    g.ndata['node_feat'] = features
    if cuda:
        norm = norm.cuda()
    g.ndata['norm'] = norm.unsqueeze(1)

    G=GraphData()
    G.from_dgl(g)    

    # create GCN model
    model = GCN(
                in_feats,
                args.n_hidden,
                n_classes,
                args.n_layers,
                args.aggregate_type,
                args.output_size,
                dropout=args.dropout)



    if cuda:
        model.cuda()
    loss_fcn = torch.nn.CrossEntropyLoss()


    # use optimizer
    optimizer = torch.optim.Adam(model.parameters(),
                                 lr=args.lr,
                                 weight_decay=args.weight_decay)



    # initialize graph
    dur = []
    for epoch in range(args.n_epochs):
        model.train()
        if epoch >= 3:
            t0 = time.time()
        # forward
        logits = model(G)
        loss = loss_fcn(logits[train_mask], labels[train_mask])

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        if epoch >= 3:
           dur.append(time.time() - t0)



        acc = evaluate(model, G, labels, val_mask)
        print("Epoch {:05d} | Time(s) {:.4f} | Loss {:.4f} | Accuracy {:.4f} | "
              "ETputs(KTEPS) {:.2f}". format(epoch, np.mean(dur), loss.item(),
                                            acc, n_edges / np.mean(dur) / 1000))



    print()
    acc = evaluate(model, G, labels, test_mask)
    print("Test accuracy {:.2%}".format(acc))





if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='GCN')
    register_data_args(parser)
    parser.add_argument("--dropout", type=float, default=0.5,
            help="dropout probability")
    parser.add_argument("--gpu", type=int, default=-1,
            help="gpu")
    parser.add_argument("--lr", type=float, default=1e-2,
            help="learning rate")
    parser.add_argument("--n-epochs", type=int, default=400,
            help="number of training epochs")
    parser.add_argument("--n-hidden", type=int, default=[64],
            help="number of hidden gcn units")
    parser.add_argument("--n-layers", type=int, default=2,
            help="number of hidden gcn layers")
    parser.add_argument("--weight-decay", type=float, default=5e-4,
            help="Weight for L2 loss")
    parser.add_argument("--self-loop", action='store_true',
            help="graph self-loop (default=False)")
    parser.add_argument('--aggregate_type', type=str, default='mean',
                        help="aggregate type: 'mean','gcn','pool','lstm'")   
    parser.add_argument("--output_size", type=int, default=16,
                        help="hiddensize")
    parser.add_argument("--direction_option", type=str, default='uni',
                        help="direction type (`uni`, `bi_fuse`, `bi_sep`)")
    parser.set_defaults(self_loop=False)
    args = parser.parse_args()
    main(args)



import torch
import torch.nn as nn
import torch.nn.functional as F
from layers import GraphConvolution
from ...modules.prediction.classification.link_prediction.ElementSumLayer import ElementSumLayer
from ...modules.prediction.classification.link_prediction.ConcatFeedForwardNNLayer import ConcatFeedForwardNNLayer
from ...modules.prediction.classification.link_prediction.StackedElementProdLayer import StackedElementProdLayer


class GCNModelVAE(nn.Module):
    def __init__(self, input_feat_dim, hidden_dim1, hidden_dim2, dropout,prediction_type):
        super(GCNModelVAE, self).__init__()
        self.gc1 = GraphConvolution(input_feat_dim, hidden_dim1, dropout, act=F.relu)
        self.gc2 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.gc3 = GraphConvolution(hidden_dim1, hidden_dim2, dropout, act=lambda x: x)
        self.prediction_type=prediction_type
        self.act=lambda x: x
        if self.prediction_type=='ele_sum':
           self.dc = ElementSumLayer(hidden_dim2,16,1)
        if self.prediction_type=='concat_NN':
           self.dc = ConcatFeedForwardNNLayer(hidden_dim2,16,1)
        if self.prediction_type=='stacked_ele_prod':
           self.dc = StackedElementProdLayer(hidden_dim2,16,1,1)     
        #InnerProductDecoder(dropout, act=lambda x: x)

    def encode(self, x, adj):
        hidden1 = self.gc1(x, adj)
        return self.gc2(hidden1, adj), self.gc3(hidden1, adj)

    def reparameterize(self, mu, logvar):
        if self.training:
            std = torch.exp(logvar)
            eps = torch.randn_like(std)
            return eps.mul(std).add_(mu)
        else:
            return mu

    def forward(self, x, adj):
        mu, logvar = self.encode(x, adj)
        z = self.reparameterize(mu, logvar)
        if self.prediction_type=='stacked_ele_prod':
                  link_logits=self.dc([z])
                  recovered=self.dc([mu])
        else:
               link_logits=self.dc(z)
               recovered=self.dc(mu)
        return self.act(link_logits), mu, logvar,recovered



class InnerProductDecoder(nn.Module):
    """Decoder for using inner product for prediction."""
    def __init__(self, dropout, act=torch.sigmoid):
        super(InnerProductDecoder, self).__init__()
        self.dropout = dropout
        self.act = act
    def forward(self, z):
        z = F.dropout(z, self.dropout, training=self.training)
        adj = self.act(torch.mm(z, z.t()))
        return adj



from torch import nn
import torch as th
import torch.autograd as autograd
from ..base import NodeClassifierLayerBase

class BiLSTMFeedForwardNNLayer(NodeClassifierLayerBase):
    r"""Specific class for node classification layer.


    Parameters
    ----------

    input_size : int 
                 The length of input node embeddings
    output_size : int 
               The number of node catrgoriey for classification
    hidden_size : int
                  the hidden size of the linear layer 

    """       
    def __init__(self, input_size, output_size,hidden_size=None,dropout=0):
        super(BiLSTMFeedForwardNNLayer, self).__init__()
        self.hidden_size=hidden_size
        self.output_size=output_size
        self.device =th.device('cuda') if th.cuda.is_available() else th.device('cpu')
        if hidden_size!=None:
          self.lstm = nn.LSTM(input_size, self.hidden_size//2, num_layers=1, bidirectional=True,batch_first=True,dropout=dropout) # Maps the output of the LSTM into tag space.                
          self.linear = nn.Linear(hidden_size, self.output_size)

        else:
          self.lstm = nn.LSTM(input_size, self.output_size//2, num_layers=1, bidirectional=True,batch_first=True,dropout=dropout) # Maps the output of the LSTM into tag space.                
         
        
        
        
    def init_hidden(self,batch_size): 
        if self.hidden_size!=None:
          return (autograd.Variable(th.randn(2, batch_size, self.hidden_size // 2)).to(self.device),                
                  autograd.Variable(th.randn(2, batch_size, self.hidden_size  // 2)).to(self.device))
        else:
          return (autograd.Variable(th.randn(2, batch_size, self.output_size // 2)).to(self.device),                
                  autograd.Variable(th.randn(2, batch_size, self.output_size // 2)).to(self.device))

    def forward(self, node_emb, node_idx=None):
        r"""
        Forward functions for classification task.
    
        ...
    
        Parameters
        ----------
    
        node_emb : padded tensor [B,N,H]
                   B: batch size
                   N: max number of nodes    
                   H: length of the node embeddings
        node_idx : a list of index of nodes that needs classification.
                   Default: 'None'
                   Example: [1,3,5]
    
        Returns 
        -------
             logit tensor: [B,N, num_class] The score logits for all nodes preidcted.
        """ 
        if node_idx:
            node_emb = node_emb[th.tensor(node_idx), :]  # get the required node embeddings.
        batch_size=node_emb.size(0) 
        
        
        self.hidden = self.init_hidden(batch_size)  
        lstm_out, self.hidden = self.lstm(node_emb, self.hidden)
        if  self.hidden_size is None:                                      
            return lstm_out  
        else:
           lstm_feats = self.linear(lstm_out)
           return lstm_feats

       
